# -*- coding: UTF-8 -*-
import datetime
import hashlib
import logging
import os
import re
import sys
import time

from autoscrape.filetypes import TEXT_EXTENSIONS
from autoscrape.util import (
    get_filename_from_url, get_extension_from_url, write_file
)


logger = logging.getLogger('AUTOSCRAPE')


class BaseScraper(object):
    """
    A base class for common scraper functionality like loglevel
    parsing & setup, file saving, etc.
    """

    def setup_logging(self, loglevel=None, stdout=False):
        if loglevel == "DEBUG":
            loglevel = logging.DEBUG
        elif not loglevel or loglevel == "INFO":
            loglevel = logging.INFO
        elif loglevel == "WARN":
            loglevel = logging.WARN
        elif loglevel == "ERROR":
            loglevel = logging.ERROR

        logger.setLevel(loglevel)
        if stdout:
            console_handler = logging.StreamHandler(stream=sys.stdout)
            logger.addHandler(console_handler)

    def save_screenshot(self, classname=None):
        """
        Save a screenshot of the current page window. The files
        created by this are quite large and also only contain the
        first, above-the-fold content.

        Screenshots are saved by the timestamp and the "class"
        of the page currently visited (data_page, etc).
        """
        if not self.output or not self.save_screenshots:
            return

        if classname not in self.training_classes:
            raise ValueError("Base class speficied: %s" % classname)

        if self.full_page_screenshots:
            # adjust the window size to the document size, this captures most
            # web configurations except cases where a footer, for example, has
            # been placed absolutely outside of the body and hangs beneath
            # NOTE: this comes from here: https://stackoverflow.com/a/52572919
            original_size = self.control.scraper.driver.get_window_size()
            required_width = self.control.scraper.driver.execute_script(
                'return document.body.parentNode.scrollWidth'
            )
            required_height = self.control.scraper.driver.execute_script(
                'return document.body.parentNode.scrollHeight'
            )
            logger.debug(
                "Required width=%s, height=%s, Original width=%s, height=%s" % (
                    required_width, required_height, original_size["width"],
                    original_size["height"]
                ))
            self.control.scraper.driver.set_window_size(
                required_width, required_height
            )

        if re.match("^https?://", self.output):
            screenshot_dir = "screenshots"
        else:
            screenshot_dir = os.path.join(self.output, "screenshots")

        filepath = os.path.join(screenshot_dir, "%s_%s.png" % (
            int(time.time()), classname
        ))

        png = None
        if hasattr(self.control.scraper, "driver"):
            logger.debug("[.] Saving screenshot to file: %s." % filepath)
            png = self.control.scraper.get_screenshot()

        # if self.control.scraper.driver_name == "Firefox":
        #     try:
        #         html_el = self.control.scraper.driver.find_element_by_tag_name(
        #             "html"
        #         )
        #         png = html_el.screenshot_as_png
        #         logger.debug("HTML root for screenshot: %s" % html_el)
        #         logger.debug("DIR: %s" % dir(html_el))
        #     except Exception as e:
        #         logger.warn("Error saving screenshot: %s" % e)

        # elif not png:
        #     try:
        #         png = self.control.scraper.driver.get_screenshot_as_png()
        #     except Exception as e:
        #         logger.warn("Error saving screenshot: %s" % e)

        if png and self.output:
            write_file(
                filepath, png, fileclass="screenshot",
                writetype="wb", output=self.output,
                url=self.control.scraper.page_url,
            )

        if self.full_page_screenshots:
            # restore original window size to avoid side effects
            self.control.scraper.driver.set_window_size(
                original_size['width'], original_size['height']
            )

    def save_training_page(self, classname=None, url=None):
        """
        Writes the current page to the output data directory (if provided)
        to the given class folder.
        """
        if not self.output and not self.return_data:
            return

        logger.debug("[.] Saving training page for class: %s" % classname)
        if classname not in self.training_classes:
            raise ValueError("Base class speficied: %s" % classname)

        data = None
        if url is None:
            # TODO: migrate over
            data = self.control.scraper.page_html
            url = self.control.scraper.page_url
        else:
            data = self.control.scraper.download_file(url)

        ext = get_extension_from_url(url)
        link_to_text = ext in TEXT_EXTENSIONS
        if not link_to_text:
            data = self.control.scraper.download_file(url, return_data=True)
            print("DOWNLOADS")
            classname = "downloads"

        # always keep filename for downloads, for now
        if not self.output or re.match("^https?://", self.output):
            classdir = classname
        else:
            classdir = os.path.join(self.output, classname)


        # we had some kind of error downloading
        if data is None:
            return

        # hash the contents of the file, so we don't *not* save dynamic
        # JS pages with the same URl and that we *don't* excessively save
        # the same page over and over
        if type(data) == bytes:
            sha256 = hashlib.sha256()
            sha256.update(data)
            h = sha256.digest().hex()
            writetype = "wb"
        else:
            h = hashlib.sha256(data.encode("utf-8")).digest().hex()
            writetype = "w"

        logger.debug("URL: %s, Hash: %s, Extension: %s" % (url, h, ext))

        # don't use the hash, use the filename from URL
        if self.keep_filename:
            parsed_filename = get_filename_from_url(url)
            logger.debug("Parsed output filename: %s" % parsed_filename)
            filepath = os.path.join(classdir, parsed_filename)
        # use the hash as the output filename
        else:
            filepath = os.path.join(classdir, "%s.%s" % (h, ext))

        crawl_data = None
        if self.output:
            logger.info(" - Saving page: %s" % (filepath))
            write_file(
                filepath, data, fileclass=classname,
                writetype=writetype, output=self.output,
                url=self.control.scraper.page_url,
            )
        if self.return_data:
            now = datetime.datetime.utcnow()
            crawl_data = {
                "url": url,
                "html": data,
                "fileclass": classname,
                "date": now.isoformat(timespec="seconds") + "Z",
            }

        # only save stylesheets for web content types
        if link_to_text and not self.disable_style_saving:
            logger.debug(" -  Saving stylesheet")
            style_filepath = "%s.css" % filepath
            # this will save stylesheet as filepath.html.css
            stylesheet = self.control.scraper.get_stylesheet()
            if self.output:
                write_file(
                    style_filepath, stylesheet,
                    fileclass=classname, output=self.output,
                    url=self.control.scraper.page_url,
                )
            if self.return_data:
                crawl_data["css"] = stylesheet

        if self.return_data:
            self.crawl_data.append(crawl_data)

    def save_scraper_graph(self):
        """
        Saves our graph that was built throughout the scrape. This can
        be used to visualize the scrape, debug it, and replicate it.

        Graph is saved to the graph subdirectory of the output
        path, with the filename a microscond timestamp with the .gpickle
        extension.

        If output is not set, then calling this function has no
        effect.
        """
        if not self.output or not self.save_graph:
            logger.debug("[!] No output or save-graph options. Not saving")
            return

        filename = "%s.gpickle" % int(time.time() * 1000)
        basedir = os.path.join(self.output, "graph")
        if not os.path.exists(basedir):
            os.makedirs(basedir)

        graph_save_path = os.path.join(basedir, filename)
        logger.debug("[.] Saving graph to: %s" % graph_save_path)
        self.control.scraper.graph.save_graph(graph_save_path)



CASES = [{
    "extractor": """
<article class="lawyer">
  <a>
    <img style:replace(/background: url\((.*)\);/, "\1"):prepend("https://texasbar.com"):profileImg />
  </a>
  <div>
    <h3>
      <a href:prepend("https://texasbar.com"):profileLink >
        <span:nth-child(1) @text:namePrefix />
        <span:nth-child(2) @text:nameFirst />
        <span:nth-child(3) @text:nameLast />
        <span:nth-child(4) @text:nameSuffix/>
      </a>
    </h3>
    <h5 @text:affiliation />
    <p:attribute-count(0) @text:practiceLocation />
    <p class="address" @text:address />
    <p class="areas" @text:practiceArea />
  </div>
</article>""",
    "html": """
<article class="lawyer">
  <div class="no-img">
    <a href="/AM/Template.cfm?Section=Find_A_Lawyer&amp;template=/Customsource/MemberDirectory/MemberDirectoryDetail.cfm&amp;ContactID=282365">
      <!--
        <img src="/AM/templates/layoutcatalog/1601/Images/ProfilePhotoPlaceholder.gif" title="State Bar Seal" alt="State Bar Seal" class="avatar"/>
      -->
      <img style="background: url(/AM/templates/layoutcatalog/1601/Images/ProfilePhotoPlaceholder.gif);" class="avatar">
    </a>
  </div>
  <div class="avatar-column">
    <!-- Determine which status class should be displayed -->
    <h3>
      <span class="status-icon blue triangle"></span>
      <a href="/AM/Template.cfm?Section=Find_A_Lawyer&amp;template=/Customsource/MemberDirectory/MemberDirectoryDetail.cfm&amp;ContactID=282365">
        <span class="honorific-prefix"></span>
        <span class="given-name">John H.</span>
        <span class="family-name">Smither</span>
        <span class="honorific-suffix"></span>
      </a>
    </h3>
    <h5>Vinson &amp; Elkins LLP</h5>
    <p>
    <strong>Primary Practice Location:</strong>
    Houston,&nbsp;TX
    </p>
    <p class="address">
    1001 Fannin St Ste 2801<br> Houston, TX&nbsp;77002
    </p>
    <p class="areas">
    <strong>Practice Areas:</strong>
    None Specified by Attorney
    </p>
    <a href="/AM/Template.cfm?Section=Find_A_Lawyer&amp;template=/Customsource/MemberDirectory/MemberDirectoryDetail.cfm&amp;ContactID=282365" class="read-more">Full profile</a>
    <!--
      <h4 class="wai">Memberships</h4>
      <ul class="memberships">
      </ul>
    -->
  </div>
  <div class="contact">
    <a href="http://h2vx.com/vcf/https://texasbar.com/AM/Template.cfm?Section=Find_A_Lawyer&amp;template=/Customsource/MemberDirectory/vcard.cfm&amp;ContactID=282365">Download vCard <i class="fa fa-download"></i></a>
    <a href="tel:713-758-2466">Tel: 713-758-2466 <i class="fa fa-phone"></i></a>
  </div>
  <div class="badges">
  </div>
</article>
    """,
}, {
    "extractor": """""",
    "html": """""",
}]



# -*- coding: UTF-8 -*-
import logging

from autoscrape.backends.base.tags import TaggerBase
from autoscrape.backends.requests.dom import Dom


logger = logging.getLogger('AUTOSCRAPE')


class Tagger(TaggerBase, Dom):
    def tag_from_element(self, el):
        path = []
        while el is not None:
            nth = 1
            parent = el.getparent()
            children = []
            if parent is not None:
                children = parent.getchildren()
            for child in children:
                if child == el:
                    break
                if child.tag == el.tag:
                    nth += 1
            selector = "%s:nth-of-type(%s)" % (
                el.tag, nth
            )
            path.insert(0, selector)
            el = parent
        tag = " > ".join(path)
        return tag

    def get_inputs(self, form=None, itype=None, root_node=None):
        return super().get_inputs(form=form, itype=itype, root_node=self.dom)

    def get_buttons(self, in_form=False, path=None):
        x_path = path or "|".join([
            "//form//a", "//input[@type='submit']", "//table//a",
        ])
        return super().get_buttons(in_form=in_form, path=x_path)

    def get_clickable(self, path=None):
        clickable = super().get_clickable(path="//a|//iframe")
        return clickable

    def clickable_sanity_check(self, element):
        raw_href = self.element_attr(element, "href")

        tag_name = self.element_tag_name(element)
        if tag_name == "iframe":
            raw_href = self.element_attr(element, "src")

        if not raw_href:
            return False

        href = self._normalize_url(raw_href).split("#")[0]
        if href.split("#")[0] == self.current_url:
            return False

        # skip any weird protos ... we whitelist notrmal HTTP,
        # anchor tags and blank tags (to support JavaScript & btns)
        if href and href.startswith("javascript"):
            return False

        return super().clickable_sanity_check(element, href=href)



# -*- coding: utf-8 -*-
import logging

from autoscrape.backends.base.tags import TaggerBase
from autoscrape.backends.selenium.dom import Dom


logger = logging.getLogger('AUTOSCRAPE')


class Tagger(TaggerBase, Dom):
    """
    Generates tags from a given page that can be used, in a stateless manner,
    to refer to unique elements on a web page.
    """

    def __init__(self, driver=None, current_url=None, leave_host=False):
        super().__init__(current_url=current_url, leave_host=leave_host)
        self.driver = driver

    def tag_from_element(self, element):
        """
        Takes a WebDriver element and returns an CSSPath for finding it
        in the future. As far as I know, this is only really feasible
        using JavaScript (without resorting to a complicated tree walking
        algorithm ... which we may need to do if this ends up failing).

        Modified from: https://stackoverflow.com/a/12222317
        """
        script = """
            var getPathTo = function(el) {
                if (!(el instanceof Element))
                    return;
                var path = [];
                while (el.nodeType === Node.ELEMENT_NODE) {
                    // if (el.id) {
                    //   path.unshift(`#${el.id}`);
                    //   break;
                    // }
                    var selector = el.nodeName.toLowerCase();
                    // // NOTE: we removed this because web pages often use
                    // // strange characters in ID names which cause the CSS
                    // // selector to fail upon lookup. If we only use traversal
                    // // methods, we don't have that webpage-specific problem
                    // if (el.id) {
                    //     selector += '#' + el.id;
                    //     path.unshift(selector);
                    //     break;
                    // }

                    var sib = el, nth = 1;
                    while (sib = sib.previousElementSibling) {
                        if (sib.nodeName.toLowerCase() == selector)
                           nth++;
                    }

                    // // NOTE: always give a nth-of-type tag, even if
                    // // if there's only a single sibling, just to be
                    // // extra-specific
                    // if (nth != 1)

                    selector += ":nth-of-type("+nth+")";
                    path.unshift(selector);
                    el = el.parentNode;
                }
                return path.join(" > ");
            }

            // NOTE: this used to have a toLowerCase on it, but it caused
            // problems with some pages. Leaving it as it was found in the
            // original DOM is best here.
            return getPathTo(arguments[0]); //.toLowerCase();
        """
        return self.driver.execute_script(script, element)

    def clickable_sanity_check(self, element):
        try:
            if not self.element_displayed(element):
                return False
        except Exception as e:
            logger.debug("[!] Skipping sanity check due to failure: %s" % (e))
            return False
        return super().clickable_sanity_check(element)

    def get_inputs(self, form=None, itype=None, root_node=None):
        return super().get_inputs(form=form, itype=itype, root_node=self.driver)

    def get_clickable(self, path=None):
        """
        Get all clickable element tags on the current page.

        TODO: In the future we may need to recurse the page to find
        other clickable types like JS-enabled divs, etc.
        """
        xpath = path or "|".join([
            "//a", "//button", "//input[@type='submit']",
            "//input[@type='button']"
        ])
        return super().get_clickable(path=xpath)



# -*- coding: utf-8 -*-
import logging
import re

try:
    from selenium.common.exceptions import (
        NoSuchElementException, StaleElementReferenceException,
        NoSuchFrameException
    )
except ModuleNotFoundError:
    # we haven't installed selenium backend deps
    pass

from autoscrape.backends.base.dom import DomBase


logger = logging.getLogger('AUTOSCRAPE')


class FrameTransparentList(list):
    def __init__(self, *args, **kwargs):
        self.driver = kwargs.pop("driver")
        super().__init__(*args, **kwargs)

    def __getitem__(self, index):
        raw_item = super().__getitem__(index)
        if not isinstance(raw_item, list):
            self.driver.switch_to.default_content()
            return raw_item
        iframe_ix, element = raw_item
        self.driver.switch_to.frame(iframe_ix)
        return element

    def __iter__(self):
        for raw_item in list.__iter__(self):
            if not isinstance(raw_item, list):
                self.driver.switch_to.default_content()
                yield raw_item
                continue
            iframe_ix, element = raw_item
            self.driver.switch_to.frame(iframe_ix)
            yield element
            self.driver.switch_to.default_content()


class Dom(DomBase):
    def element_attr(self, element, name, default=None):
        return element.get_attribute(name)

    def iframe_capable_lookup(self, tag):
        try:
            self.driver.switch_to.default_content()
            return self.driver.find_element_by_css_selector(tag)
        except NoSuchElementException:
            pass
        self.driver.switch_to.default_content()
        iframes = self.driver.find_elements_by_tag_name("iframe")
        for iframe_ix in range(len(iframes)):
            try:
                self.driver.switch_to.frame(iframe_ix)
            except NoSuchFrameException:
                continue
            try:
                return self.driver.find_element_by_css_selector(tag)
            except NoSuchElementException:
                continue
            self.driver.switch_to.default_content()
        logger.debug("[!] No element found for tag: %s" % (tag))
        return None

    def element_by_tag(self, tag):
        """
        Take a tag and return the corresponding live element in the DOM.
        """
        inside_id = False
        # escaping logic
        newtag = ""
        for c in tag:
            if c == "#":
                inside_id = True
                newtag += c
                continue

            # end of ID
            elif inside_id and re.search("\s", c):
                inside_id = False

            elif inside_id and c in self.css_escapables:
                for escapable in self.css_escapables:
                    c = "\%s" % escapable

            newtag += c

        if newtag != tag:
            logger.debug("Original tag: %s, newtag: %s" % (tag, newtag))
            tag = newtag

        return self.iframe_capable_lookup(tag)

    def elements_by_path(self, xpath, from_element=None):
        """
        Get all elements, across all iframes. We output a
        FrameTransparentList which is a normal list of elements, but it will
        switch to the correct frame upon accessing/iterating to
        each element.
        """
        if from_element is None:
            from_element = self.driver
        return from_element.find_elements_by_xpath(xpath)
        iframes = from_element.find_elements_by_tag_name("iframe")
        if not len(iframes):
            return from_element.find_elements_by_xpath(xpath)
        # gather all elements from iframe
        elements = from_element.find_elements_by_xpath(xpath)
        for iframe_ix in range(len(iframes)):
            self.driver.switch_to.frame(iframe_ix)
            for el in self.driver.find_elements_by_xpath(xpath):
                elements.append([iframe_ix, el])
            self.driver.switch_to.default_content()
        self.driver.switch_to.default_content()
        return FrameTransparentList(elements, driver=self.driver)

    def get_stylesheet(self):
        script = """
        return [].slice.call(document.styleSheets)
          .reduce((prev, styleSheet) => {
            try {
              if (styleSheet.cssRules) {
                return prev +
                  [].slice.call(styleSheet.cssRules)
                    .reduce(function (prev, cssRule) {
                      return prev + cssRule.cssText;
                    }, '');
              } else {
                  return prev;
              }
            } catch (e) {
              return prev + `@import url("${styleSheet.href}");`
            }
          }, '');"""
        return self.driver.execute_script(script)

    def _text_via_many_means(self, el):
        text = []
        try:
            txt = el.text
            if txt:
                text.append(txt.strip())
        except Exception as e:
            logger.error("Error getting text element: %s, Err: %s" % (
                el, e))

        title = el.get_attribute("title")
        if title:
            text.append(title.strip())

        try:
            placeholder = el.get_attribute("placeholder")
            if placeholder:
                text.append(placeholder.strip())
        except Exception as e:
            logger.error("Error getting placeholder: %s, Error: %s" % (
                el, e))

        img_els = el.find_elements_by_tag_name("img")
        for img in img_els:
            try:
                text.append(img.get_attribute("alt"))
            except StaleElementReferenceException as e:
                logger.error("Error getting image text: %s, Error: %s" % (
                    img, e
                ))

        if self.element_tag_name(el) == "input":
            text.append(el.get_attribute("value"))

        return " ".join(text).replace("\n", "").strip()

    def element_text(self, element, block=False):
        """
        Get the text for all elements either under a given element
        or for a whole page (if element == None)
        """
        if not block and element is not None:
            return self._text_via_many_means(element)

        if element is None:
            element = self.driver

        return element.text

    def element_tag_name(self, element):
        if element is None:
            return ""
        return element.tag_name

    def element_value(self, element):
        return element.get_attribute("value")

    def element_displayed(self, element):
        fn_names = ["is_displayed", "is_enabled"]
        for fn_name in fn_names:
            if not hasattr(element, fn_name):
                continue
            try:
                if not getattr(element, fn_name)():
                    return False
            except StaleElementReferenceException as e:
                pass
        return True



# -*- coding: UTF-8 -*-
import io
import logging
import os
import pickle
import sys

from autoscrape.backends.requests.browser import RequestsBrowser
from autoscrape.backends.requests.tags import Tagger
from autoscrape.search.graph import Graph
from autoscrape.util.warc import build_warc_index, _warc_records


logger = logging.getLogger('AUTOSCRAPE')


try:
    import plyvel
    import warcio
except ModuleNotFoundError:
    pass


class WARCBrowser(RequestsBrowser):
    def __init__(self, warc_index_file=None, warc_directory=None,
                 filter_domain=None, leave_host=False, **kwargs):
        try:
            warcio
        except NameError:
            logger.debug(
                "WARC dependencies not installed."
                " (Hint: pip install autoscrape[warc-backend])"
                " Exiting."
            )
            sys.exit(1)

        no_dir_msg = "Error: No warc_directory specified for WARCBrowser"
        assert warc_directory is not None, no_dir_msg

        no_index_msg = "Error: No warc_index_file specified for WARCBrowser"
        assert warc_index_file is not None, no_index_msg

        # leveldb directory
        self.warc_index_file = warc_index_file
        # directory containing Common Crawl WARCs
        self.warc_directory = warc_directory
        # only build index for a specific domain
        self.filter_domain = filter_domain

        # WARC index: URL => (filename, record_number)
        self.warc_index = plyvel.DB(self.warc_index_file, create_if_missing=True)
        build_warc_index(
            db=self.warc_index, warc_directory=self.warc_directory,
            filter_domain=self.filter_domain
        )
        # WARC cache: filename => [record1, ..., recordN]
        self.warc_cache = {}
        self.warc_directory = warc_directory

        # how many WARC files to keep in memory at a given time
        # since the crawls are sequential, most files for a site
        # will exist in a segment of a few WARC files.
        self.warc_cache_size = 2
        # we're going to store the order the files have have been
        # accessed most recently here:
        #     [most_recently_used_filename, ..., least_recently_used_filename]
        # This will be used to enforce our cache size.
        self.warc_cache_stack = []

        # set of clicked elements
        self.visited = set()

        # queue of the path that led us to the current page
        # this is in the form of (command, *args, **kwargs)
        self.path = []

        # tree building
        self.graph = Graph()

        # setting to False, ensures crawl will stay on same host
        self.leave_host = leave_host

        self.current_url = None
        self.current_html = None

    def _load_warc_file(self, filename):
        """
        Take a specified WARC file, load it and keep it in memory in a quickly
        readable format (python dict). This operates directly on the class
        variable self.warc_cache and also handles maximum cache size pruning.
        """
        logger.debug("[-] Loading WARC file: %s" % (filename))
        if len(self.warc_cache_stack) > self.warc_cache_size:
            least_used = self.warc_cache_stack.pop()
            logger.debug(" - Removing WARC from memory: %s" % (filename))
            del self.warc_cache[least_used]

        self.warc_cache[filename] = []
        for record in _warc_records(filename):
            payload = record["payload"]
            if not payload:
                payload = "<html></html>"
            self.warc_cache[filename].append({
                "header": record["headers"],
                "payload": payload,
            })

    def fetch(self, url, initial=False):
        """
        Fetch a page from a given URL from the WARC archive (via
        an index).
        """
        logger.info("%s Fetching url=%s initial=%s" % (
            ("[+]" if initial else " -"), url, initial,
        ))
        url_b = bytes(url, "utf-8")
        data = self.warc_index.get(url_b)
        if not data:
            logger.debug("[!] Couldn't find URL in WARC index: %s" % (url))
            return False
        else:
            filename, record_number = pickle.loads(data)
            logger.debug(" -  Loading filename: %s record number: %s" % (
                filename, record_number
            ))
            if filename not in self.warc_cache:
                self._load_warc_file(filename)
            warcfile = self.warc_cache[filename]
            record = warcfile[record_number]
            self.current_html = record["payload"]

            try:
                self.warc_cache_stack.remove(filename)
            except ValueError:
                pass

            self.warc_cache_stack.insert(0, filename)

        self.current_url = url
        self.dom = self._get_dom()

        if initial:
            self.path.append(("fetch", [url], {"initial": initial}))
            node = "Fetch\n url: %s" % url
            self.graph.add_root_node(node, url=url, action="fetch")

        return True



# -*- coding: UTF-8 -*-
import logging
import re

from autoscrape.scrapers import BaseScraper
from autoscrape.control import Controller
from autoscrape.input_parser import InputParser


logger = logging.getLogger('AUTOSCRAPE')


class ManualControlScraper(BaseScraper):
    """
    A Depth-First Search scraper that looks for forms, inputs, and next
    buttons by some manual criteria and iterates accordingly.
    """
    training_classes = [
        "data_pages", "error_pages", "links_to_documents",
        "links_to_search", "search_pages", "crawl_pages",
        "interaction_pages",
    ]

    # TODO: save the path to the matched search form to a file. then
    # upon subsequent loads, we can try that path first and then go
    # from there. this way the manual control scraper can learn from
    # self-exploration based on manual params
    #
    # Another related idea: it would be nice if we had an API that
    # saved data about successful runs and loading/replaying them
    # so users can get familiar with the concepts of self-exploration
    # and self-learning but without having to get the ML concepts.

    def __init__(self, baseurl, loglevel=None, stdout=None, maxdepth=10,
                 max_pages=None,
                 formdepth=0, next_match=None, form_match=None,
                 output=None, keep_filename=False, disable_style_saving=False,
                 save_screenshots=False, save_graph=False,
                 full_page_screenshots=False, input=None, leave_host=False,
                 driver="Firefox", browser_binary=None,
                 remote_hub="http://localhost:4444/wd/hub",
                 link_priority=None, ignore_links=None, only_links=None,
                 ignore_extensions=None, result_page_links=None,
                 form_submit_natural_click=False, form_submit_wait=5,
                 force_page_wait=None, form_submit_button_selector=None,
                 load_images=False, show_browser=False, warc_index_file=None,
                 warc_directory=None, return_data=False, page_timeout=None,
                 backend="selenium"):
        # setup logging, etc
        super().setup_logging(
            loglevel=loglevel, stdout=stdout
        )
        if page_timeout is not None:
            page_timeout = int(page_timeout or 0)
        # set up web scraper controller
        self.control = Controller(
            leave_host=leave_host, driver=driver, remote_hub=remote_hub,
            form_submit_natural_click=form_submit_natural_click,
            form_submit_wait=int(form_submit_wait or 0),
            form_submit_button_selector=form_submit_button_selector,
            force_page_wait=int(force_page_wait or 0),
            load_images=load_images, show_browser=show_browser,
            warc_index_file=warc_index_file, warc_directory=warc_directory,
            output=output, backend=backend, browser_binary=browser_binary,
            page_timeout=page_timeout,
        )
        self.control.initialize(baseurl)
        # depth of DFS in search of form
        self.maxdepth = int(maxdepth or 0)
        # current depth of iterating through 'next' form buttons
        self.formdepth = int(formdepth or 0)
        # max total pages to fetch
        self.max_pages = int(max_pages) if max_pages else None
        # keep number of fetched pages here.
        self.total_pages = 0
        # match for link to identify a "next" button
        self.next_match = next_match
        # string to match a form (by element text) we want to scrape
        self.form_match = form_match
        # Where to write training data from crawl
        self.output = output
        # Whether or not to return crawled data upon completion. This can be
        # used along with output, or on its own. This will store all
        # data in memory (in self.crawl_data) during the crawl, so beware!
        self.return_data = return_data
        self.crawl_data = []
        # If this is true, do not use a file hash for the filename
        self.keep_filename = keep_filename
        # Disable saving of stylesheets for web content types
        self.disable_style_saving = disable_style_saving
        # To save screenshots or not (they're large and expensive)
        self.save_screenshots = save_screenshots
        # To save the whole, scrolled down page screenshot, this can
        # fail in certain circumstances
        self.full_page_screenshots = full_page_screenshots
        # Whether to save the crawl & interaction traversal graph
        self.save_graph = save_graph
        # string used to match link text in order to sort them higher
        self.link_priority = link_priority
        # string or regex to be used to omit links from clickable
        self.ignore_links = ignore_links
        # string or regex to use to not go to URLs with extensions
        self.ignore_extensions = ignore_extensions
        # Whitelisted links to click
        self.only_links = only_links
        # Apply any link clicking rules to the results pages
        self.result_page_links = result_page_links
        # attempt a position-based "natural click" over the element
        self.form_submit_natural_click = form_submit_natural_click
        # a period of seconds to force a wait after a submit
        self.form_submit_wait = int(form_submit_wait or 0)
        # a generator, outputting individual form interaction plans
        if input:
            self.input_gen = InputParser(input).generate()
        # if not specified, do nothing with forms
        else:
            self.input_gen = [[]]
        # whether or not we've successfully scraped what we want
        self.scraped = False

    def click_until_no_links(self, links):
        logger.info("[.] Clicking result page links...")
        if self.max_pages is not None and self.total_pages >= self.max_pages:
            logger.info(" - Maximum pages %s reached, returning..." % self.max_pages)
            return

        link_vectors = self.control.vectorizer.link_vectors()
        link_zip = list(zip(range(len(link_vectors)), link_vectors))
        link_zip = filter(
            lambda x: re.findall(self.result_page_links, x[1]),
            link_zip
        )
        logger.debug(" - Candidate links: %s" % (link_zip))
        # Click until we get no more matches
        for ix, text in link_zip:
            logger.info("[.] Clicking result page link: %s" % (text))
            logger.debug(" - Current URL: %s" % (self.control.scraper.page_url))
            if self.control.select_link(ix, iterating_form=True):
                self.total_pages += 1
                self.click_until_no_links(links)
                self.save_training_page(classname="data_pages")
                self.save_screenshot(classname="data_pages")
                self.control.back()

    def keep_clicking_next_btns(self):
        """
        This looks for "next" buttons, or (in the future) page number
        links, and clicks them until one is not found. This saves the
        pages as it goes.
        """
        logger.info("[*] Entering result page iteration routine")
        depth = 0
        while True:
            if self.formdepth and depth >= self.formdepth:
                logger.debug("[*] Max 'next' formdepth reached %s" % depth)
                break

            button_data = self.control.vectorizer.button_vectors()
            n_buttons = len(button_data)
            logger.info("[.] On result page %s" % (depth + 1))
            logger.debug(" - Button vectors (%s): %s" % (
                n_buttons, button_data
            ))

            if self.result_page_links:
                self.click_until_no_links(self.result_page_links)

            # element type, index, text
            next_found = None
            for ix in range(n_buttons):
                button_text = button_data[ix]
                logger.debug(" - Checking button: %s" % button_text)
                if re.findall(self.next_match.lower(), button_text.lower()):
                    next_found = ("button", ix, button_text)
                    break

            if not next_found:
                link_vectors = self.control.vectorizer.link_vectors()
                n_clickable = len(link_vectors)
                logger.debug(" - Button not found, searching %s links" % (
                    n_clickable
                ))
                for ix in range(n_clickable):
                    link_text = link_vectors[ix]
                    logger.debug(" - Checking clickable: %s" % link_text)
                    if re.findall(self.next_match.lower(), link_text.lower()):
                        next_found = ("link", ix, link_text)
                        break

            # we didn't find a next match, break loop
            if next_found is None:
                logger.debug(" - Next button not found!")
                break
            else:
                ntype, ix, text = next_found
                logger.info("[.] Next button found! Clicking: %s" % (text))
                depth += 1
                if ntype == "button":
                    self.control.select_button(ix, iterating_form=True)
                elif ntype == "link":
                    self.control.select_link(ix, iterating_form=True)

                # subsequent page loads get saved here
                self.save_training_page(classname="data_pages")
                self.save_screenshot(classname="data_pages")

            # check for infinite loop, this is based on the hash
            # of the previous few pages
            logger.debug(" - Checking for infinite loop...")
            if self.control.scraper.infinite_loop_detected:
                logger.debug(" - Infinte loop detected. Breaking.")
                break

        for _ in range(depth):
            logger.debug("[.] Going back from result page...")
            self.control.back()

    def scrape(self, depth=0):
        logger.info("[.] Crawl depth %s" % depth)
        logger.info(" - Total pages: %s of max: %s" % (
            self.total_pages, self.max_pages
        ))
        if self.maxdepth != -1 and depth > self.maxdepth:
            logger.info(" - Maximum depth %s reached, returning..." % depth)
            self.control.back()
            return
        if self.max_pages is not None and self.total_pages >= self.max_pages:
            logger.info(" - Maximum pages %s reached, returning..." % self.max_pages)
            return
        if self.scraped:
            logger.debug(" - Scrape complete, not clicking anything else.")
            return

        if self.ignore_extensions and re.findall(self.ignore_extensions,
                                                 self.control.scraper.page_url):
            logger.debug(" - Ignoring URL matching ignored extension: %s" % (
                self.control.scraper.page_url
            ))
            return

        self.save_training_page(classname="crawl_pages")
        self.save_screenshot(classname="crawl_pages")
        form_vectors = self.control.vectorizer.form_vectors()

        # NOTE: we never get into this loop if self.input_gen is empty
        # this arises when input was not handed to the initializer
        for ix in range(len(form_vectors)):
            # don't bother with looking for forms if we didn't specify
            # th form_match option
            if not self.form_match:
                continue

            form_data = form_vectors[ix]

            # inputs are keyed by form index, purely here for debug purposes
            inputs = self.control.inputs[ix]
            logger.debug(" - Form: %s Text: %s" % (ix, form_data))
            logger.debug(" - Inputs: %s" % inputs)

            if self.form_match.lower() not in form_data.lower():
                continue

            logger.info("[*] Found an input form (No. %s on page)" % (ix))
            self.save_training_page(classname="search_pages")
            self.save_screenshot(classname="search_pages")

            for input_phase in self.input_gen:
                logger.debug(" - Input plan: %s" % input_phase)
                for single_input in input_phase:
                    input_index = single_input["index"]
                    if single_input["type"] == "input":
                        input_string = single_input["string"]
                        logger.info("[.] Inputting %s to input %s" % (
                            input_string, ix
                        ))
                        self.control.input(ix, input_index, input_string)
                    elif single_input["type"] == "select":
                        input_string = single_input["string"]
                        logger.info("[.] Selecting option %s in input %s" % (
                            input_string, input_index
                        ))
                        self.control.input_select_option(
                            ix, input_index, input_string
                        )
                    elif single_input["type"] == "checkbox":
                        to_check = single_input["action"]
                        logger.info("[.] %s checkbox input %s" % (
                            "Checking" if to_check else "Unchecking",
                            input_index
                        ))
                        self.control.input_checkbox(
                            ix, input_index, to_check
                        )
                    elif single_input["type"] == "date":
                        input_string = single_input["string"]
                        logger.info("[.] Setting date to %s in date input %s" % (
                            input_string, ix))
                        self.control.input_date(ix, input_index, input_string)
                    elif single_input["type"] == "radio":
                        radio_index = single_input["string"]
                        logger.info("[.] Selecting radio checkbox %s in group %s" % (
                            radio_index, input_index
                        ))
                        self.control.input_radio_option(
                            ix, input_index, radio_index
                        )

                # capture post-input screenshot
                self.save_screenshot(classname="interaction_pages")

                # actually submit the page
                self.control.submit(ix)
                self.total_pages += 1

                # save the initial landing result page
                self.save_screenshot(classname="data_pages")
                self.save_training_page(classname="data_pages")

                # if we're looking for next buttons, click them
                if self.next_match:
                    self.keep_clicking_next_btns()

                self.scraped = True
                self.control.back()

            logger.debug("[*] Completed iteration!")
            # Only scrape a single form, due to explicit, single
            # match configuration option
            if self.scraped:
                logger.info("[*] Scrape complete! Exiting.")
                return

        link_vectors = self.control.vectorizer.link_vectors()
        logger.debug("[.] Links on page: %s" % (link_vectors))
        link_zip = list(zip(range(len(link_vectors)), link_vectors))
        if self.ignore_links:
            logger.debug(" - Ignoring links matching: %s" % self.ignore_links)
            link_zip = filter(
                lambda x: not re.findall(self.ignore_links, x[1]),
                link_zip
            )
        if self.only_links:
            logger.debug(" - Keeping only links matching: %s" % self.ignore_links)
            link_zip = filter(
                lambda x: re.findall(self.only_links, x[1]),
                link_zip
            )
        if self.link_priority:
            logger.debug(" - Sorting by link priority: %s" % self.link_priority)
            link_zip.sort(
                key=lambda x: not re.findall(self.link_priority, x[1])
            )

        for ix, text in link_zip:
            logger.debug(" - Link index: %s text: %s" % (ix, text))
            if self.maxdepth != -1 and depth == self.maxdepth:
                logger.debug(" - At maximum depth: %s, skipping links." % depth)
                break
            if self.max_pages is not None and self.total_pages >= self.max_pages:
                logger.info(" - Maximum pages reached, skipping links.")
                break
            if self.scraped:
                logger.debug(" - Scrape complete, not clicking anything else.")
                return

            logger.debug(" - Current URL: %s" % (self.control.scraper.page_url))
            logger.debug(" - Attempting to click link text: %s" % text)
            if self.control.select_link(ix):
                logger.info("[.] Link clicked: %s" % (text))
                logger.debug(" - Current URL: %s" % (self.control.scraper.page_url))
                self.total_pages += 1
                self.scrape(depth=depth + 1)
            else:
                logger.debug(" - Click failed, skipping: %s" % text)

        logger.debug("[*] Searching forms and links on page complete")
        self.control.back()

    def run(self, *args, **kwargs):
        # we have to catch this so, in the case of failure, we
        # don't have random browser windows hanging around
        try:
            self.scrape(*args, **kwargs)
        except Exception as e:
            msg = "[!] Fatal error scraping: %s. Cleaning up, quitting."
            logger.error(msg % (e))
            if hasattr(self.control.scraper, "driver"):
                self.control.scraper.driver.quit()
            if self.output and self.save_graph:
                self.save_scraper_graph()
            raise e
        # else:
        #     logger.info("[+] AutoScrape run complete.")
        #     if self.output and self.save_graph:
        #         self.save_scraper_graph()
        try:
            self.control.scraper.driver.quit()
        except Exception:
            pass

        if self.return_data:
            return self.crawl_data



# -*- coding: UTF-8 -*-
import logging
import os
import re
import urllib

from autoscrape.util import write_file, get_filename_from_url


logger = logging.getLogger('AUTOSCRAPE')


class DomBase:
    """
    Stateful base of a web scraper. This class deals with finding and interacting
    with elements and tags. It also holds the base state variables like
    current url.
    """

    def __init__(self, leave_host=False, current_url=None, current_html=None):
        self.leave_host = leave_host
        self.current_url = current_url
        self.current_html = current_html

    def elements_by_path(self, path, from_element=None):
        """
        Return element nodes matching a path where path could be xpath,
        css, etc, depending on the backend)
        """
        raise NotImplementedError("DomBase.elements_by_path not implemented")

    def element_attr(self, element, name, default=None):
        """
        For a given element and attribute name, return the value if it
        exists.
        """
        raise NotImplementedError("DomBase.element_attr not implemented")

    def element_by_tag(self, tag):
        """
        For a given tag, return the specified element.
        """
        raise NotImplementedError("DomBase.element_by_tag not implemented")

    def get_stylesheet(self):
        """
        Return the text of all loaded CSS stylesheets.
        """
        raise NotImplementedError("DomBase.get_stylesheet not implemented")

    def element_tag_name(self):
        """
        Return the tag name of the given element.
        """
        raise NotImplementedError("DomBase.element_tag_name not implemented")

    def element_text(self, element, block=False):
        """
        Return the text of an element, or the combined text of all its
        descendants (if block=True).
        """
        raise NotImplementedError("DomBase.element_text not implemented")

    def element_value(self, element):
        """
        Return the text value of an element (e.g., input element). Since
        this is usually called like element.value() or element.value, we
        wrap this functionality here.
        """
        if not hasattr(element, "value"):
            raise NotImplementedError("DomBase.element_value not implemented")
        return element.value

    def element_name(self, element):
        """
        Return the name of an element (e.g., input element).        """
        if not hasattr(element, "name"):
            raise NotImplementedError("DomBase.element_name not implemented")
        return element.name

    def element_displayed(self, element):
        """
        Figure out if an element is displayed. If its not, we can't really
        interact with it (get its text or attribs, etc). This only effects
        the selenium backend right now.
        """
        return True

    def download_file(self, url, return_data=False):
        """
        Fetch the given url, returning a byte stream of the page data. This
        really is only useful in situations where the scraper is on a binary
        filetype, such as PDF, etc.

        Note that we're doing this as opposed to some XHR thing inside the
        selenium driver due to CORS issues.
        """
        logger.debug("Fetching non-HTML page directly: %s" % url)
        user_agent = (
            "Mozilla/5.0 "
            "(Windows NT 10.0; Win64; x64; rv:62.0) "
            "Gecko/20100101 Firefox/62.0"
        )
        request = urllib.request.Request(url, headers={
            "User-Agent": user_agent,
            "Referrer": self.page_url,
        })

        try:
            response = urllib.request.urlopen(request)
        except urllib.error.HTTPError as e:
            logger.debug("[!] HTTP error while downloading: %s" % (e))
            return

        data = response.read()
        action = {
            "action": "download_file",
            "url": url,
        }
        self.graph.add_action_to_current(action)
        if return_data:
            return data

        # always keep filename for downloads, for now
        if re.match("^https?://", self.output):
            dl_dir = "downloads"
        else:
            dl_dir = os.path.join(self.output, "downloads")

        parsed_filename = get_filename_from_url(url)
        logger.debug("Parsed output filename: %s" % parsed_filename)
        filepath = os.path.join(dl_dir, parsed_filename)
        write_file(
            filepath, data, fileclass="download", writetype="wb",
            output=self.output, url=self.page_url,
        )

    def _no_tags(self, list, l_type="path"):
        clean = []
        for p in list:
            name, args, kwargs = p
            if name == "click":
                args[0] = "tag"
            clean.append((name, args, kwargs))
        return clean



import unittest

from autoscrape.input_parser import InputParser


class TestInputParser(unittest.TestCase):
    def test_normal_string_parsing(self):
        inp_str = "i:0:string%,c:1:True,s:2:France"
        inputs_gen = InputParser(inp_str).generate()
        inputs = list(inputs_gen)
        self.assertEqual(len(inputs), 1)
        first_search = inputs[0]
        self.assertEqual(len(first_search), 3)

    def test_normal_string_multi_search_parsing(self):
        inp_str = "i:0:string1%,s:2:France;i:0:string2%,s:2:Russia"
        inputs_gen = InputParser(inp_str).generate()
        inputs = list(inputs_gen)
        self.assertEqual(len(inputs), 2)
        search = inputs[0]
        self.assertEqual(len(search), 2)
        self.assertEqual(search[0]["string"], "string1%")
        self.assertEqual(search[0]["type"], "input")
        self.assertEqual(search[0]["index"], 0)
        self.assertEqual(search[1]["string"], "France")
        self.assertEqual(search[1]["type"], "select")
        self.assertEqual(search[1]["index"], 2)
        search = inputs[1]
        self.assertEqual(len(search), 2)
        self.assertEqual(search[0]["string"], "string2%")
        self.assertEqual(search[0]["type"], "input")
        self.assertEqual(search[0]["index"], 0)
        self.assertEqual(search[1]["string"], "Russia")
        self.assertEqual(search[1]["type"], "select")
        self.assertEqual(search[1]["index"], 2)


if __name__ == "__main__":
    unittest.main()



# -*- coding: UTF-8 -*-
import logging
import re
import time

import requests
from autoscrape.backends.base.browser import BrowserBase
from autoscrape.backends.requests.tags import Tagger
from autoscrape.search.graph import Graph


logger = logging.getLogger('AUTOSCRAPE')


class RequestsBrowser(BrowserBase, Tagger):
    """
    A simple HTTP-requests based scraper, currently capable of only
    doing crawls and very basic HTTP Post requests, but is between
    2x and 5x faster.

        bxroberts.org full crawl - requests backend
        real    0m34.066s
        user    0m23.062s
        sys     0m0.640s

        bxroberts.org full crawl - selenium backend
        real    2m5.373s
        user    1m0.872s
        sys     0m10.976s
    """

    def __init__(self, leave_host=False, **kwargs):
        # requests Session
        self.s = requests.Session()
        self.s.headers.update({
            "User-Agent": (
                "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:35.0)"
                "Gecko/20100101 Firefox/35.0"
            )
        })

        # set of clicked elements
        self.visited = set()

        # queue of the path that led us to the current page
        # this is in the form of (command, *args, **kwargs)
        self.path = []

        # tree building
        self.graph = Graph()

        # setting to False, ensures crawl will stay on same host
        self.leave_host = leave_host

        self.current_url = None
        self.current_html = None

    def _no_endslash_url(self, url):
        # remove trailing slash
        url_noslash = re.sub(r"/$", "", url)
        # and trailing slash before query params
        return re.sub(r"/&", "&", url_noslash)

    def _check_and_set_visited(self, url):
        """
        Take a URL and see if we've visited it (or a slash-ending
        variation) if we haven't this sets the URL to the
        visited set. Returns True if we've visited False if not.
        """
        url_no_endslash = self._no_endslash_url(url)
        if url_no_endslash in self.visited or url in self.visited:
            return True
        self.visited.add(url)
        self.visited.add(url_no_endslash)
        return False

    def click(self, tag, **kwargs):
        element = self.element_by_tag(tag)
        text = self.element_text(element)
        url = None
        tag_name = self.element_tag_name(element)
        logger.debug(" - tag name: %s Text: %s" % (tag_name, text))
        if tag_name == "a":
            raw_href = self.element_attr(element, "href")
            if not raw_href:
                logger.debug("[!] No HREF, skipping link...")
                return False

            url = self._normalize_url(raw_href)
            if self._check_and_set_visited(url):
                logger.debug("[!] Already visited URL %s" % (url))
                return False

            logger.info("[+] Clicking link: %s" % url)
            if not self.fetch(url):
                logger.debug("[!] Fetch failed on %s" % (url))
                return False
        elif tag_name == "input":
            element_type = element.type
            if element_type == "submit":
                parent_form = element.xpath(".//ancestor::form")[0]
                parent_form_tag = self.tag_from_element(parent_form)
                self.submit(parent_form_tag, add_node=False)
                url = self.current_url
                if self._check_and_set_visited(url):
                    return False

        elif tag_name == "iframe":
            raw_href = self.element_attr(element, "src")
            if not raw_href:
                return False

            url = self._normalize_url(raw_href)
            if self._check_and_set_visited(url):
                return False

            logger.info("[+] Fetching iframe: %s" % url)
            if not self.fetch(url):
                return False
        else:
            raise NotImplementedError(
                "click not implemented for element: %s" % (tag_name)
            )

        self.path.append((
            "click", [tag], {"url": url}
        ))
        node = "Click\n text: %s\n URL: %s" % (text, url)
        node_meta = {
            "click": tag,
            "click_text": text,
            "click_iterating_form": None,
        }
        self.graph.add_node(
            node,
            **node_meta
        )
        self.graph.move_to_node(node)
        return True

    def fetch(self, url, initial=False):
        """
        Fetch a page from a given URL (entry point, typically). Most of
        the time we just want to click a link or submit a form using
        webdriver.
        """
        logger.info("%s Fetching url=%s initial=%s" % (
            ("[+]" if initial else " -"), url, initial,
        ))
        retries = 3
        success = True
        while True:
            try:
                response = self.s.get(url)
                break
            except requests.exceptions.ConnectionError as e:
                logger.error(" ! Connection error retrying...")
                logger.error(e)
                if not retries:
                    logger.error(" ! Connection error, skipping URL...")
                    return False
                time.sleep(30)
            retries -= 1

        if not response.text:
            logger.error(" ! Blank response. Skipping URL...")
            return False

        # Requests' encoding detection is faulty. The following
        # block will fix most issues
        if response.encoding and "utf" not in response.encoding.lower():
            response.encoding = response.apparent_encoding
        self.current_html = response.text
        # this check fixes improper decoding of UTF byte order mark
        if self.current_html[:3] == "ï»¿":
            self.current_html = self.current_html.encode(
                response.encoding
            ).decode("utf-8-sig")

        self.current_url = response.url
        self.dom = self._get_dom()

        if initial:
            self.path.append(("fetch", [url], {"initial": initial}))
            node = "Fetch\n url: %s" % url
            self.graph.add_root_node(node, url=url, action="fetch")

        return True

    def back(self):
        logger.info("[+] Going back...")
        logger.debug(" - current path-length=%s path=%s" % (
            len(self.path), self._no_tags(self.path),
        ))
        # We're now where we started from
        self.path.pop()
        if not self.path:
            self.path = []
            return

        prev = self.path[-1]
        if prev[0] == "fetch":
            self.graph.move_to_parent()
            self.fetch(prev[1][0])

        elif prev[0] == "click":
            self.graph.move_to_parent()
            self.fetch(prev[2]["url"])

    @property
    def page_html(self):
        return self.current_html

    @property
    def page_url(self):
        return self.current_url

    def get_clickable(self, **kwargs):
        logger.debug(" - Getting clickable...")
        self.dom = self._get_dom()
        tagger = Tagger(
            current_html=self.current_html,
            current_url=self.current_url,
            leave_host=self.leave_host,
        )
        clickable = tagger.get_clickable()
        return clickable

    # def download_file(self, url):
    #     response = self.s.get(url)
    #     action = {
    #         "action": "download_page",
    #         "url": url,
    #     }
    #     self.graph.add_action_to_current(action)
    #     return response.text

    def input(self, tag, input):
        """
        Enter some input into an element by a given tag.
        """
        logger.info("[+] Injecting text \"%s\" to input" % (input))
        elem = self.element_by_tag(tag)

        input_name = self.element_name(elem)
        value = self.element_value(elem)
        logger.debug("Input name=%s value=%s" % (input_name, value))

        elem.attrib["value"] = input

        self.path.append(("input", ("", input,), {}))
        action = {
            "action": "input",
            "text": input,
            "tag": tag,
        }
        self.graph.add_action_to_current(action)

    def submit(self, tag, add_node=True):
        """
        Submit a form, by extracting the method and url, then
        constructing the params and sending the request, along
        with the form data.
        """
        form = self.element_by_tag(tag)
        inputs = self.elements_by_path("//input", from_element=form)

        data = {}
        for input in inputs:
            # TODO: handle non-text type inputs
            name = self.element_name(input)
            if not name:
                continue
            value = self.element_value(input)
            data[name] = value

        action = self.element_attr(form, "action", default=self.current_url)
        method = self.element_attr(form, "method", default="get")
        url = self._normalize_url(action)

        request_kwargs = {}
        if method.lower() == "post":
            request_kwargs["data"] = data
        elif method.lower() == "get":
            params = []
            for key in data.keys():
                params.append([key, data.get(key, "")])
            request_kwargs["params"] = params

        request = requests.Request(
            method,
            url,
            **request_kwargs
        )
        prepped = request.prepare()
        response = self.s.send(prepped)
        self.current_url = response.url
        self.current_html = response.text
        self.dom = self._get_dom()

        # TODO: all higher level stuff
        if add_node:
            self.path.append(("submit", (tag,), {}))
            node = "Submit\n tag: %s" % (tag)
            node_meta = {
                "submit": tag,
            }
            self.graph.add_node(node, **node_meta)
            self.graph.move_to_node(node)

    def get_screenshot(self):
        return None



# -*- coding: UTF-8 -*-
# flake8: noqa: F401
__title__ = 'autoscrape-py'
__author__ = 'Brandon Roberts (brandon@bxroberts.org)'
__license__ = 'AGPLv3'
__version__ = '1.6.13'


from autoscrape.scrapers.test import TestScraper
from autoscrape.scrapers.null import NullScraper
from autoscrape.scrapers.manual import ManualControlScraper


"""
COMMAND        Logical Control Flow Step
--------       ---------------------------------------------------------------
INIT (url)               initialize & get entry point
                                     │
                                     ↓
                                 load page    🠤───────────────────┐
                                     │                            │
GET_CLICKABLE                        │        click a link based on likelihood
SELECT_LINK (index)                  │               of finding a search form
                                     ↓                            │
GET_FORMS    ┌────🠦 look for search form (possibly classifier) ───┘
             │                       │
             │                       │ FOUND
             │                       ↓
GET_INPUTS   │         identify forms on page that require input
             │     (begin with config then move to heuristic then ML)
             │                       │
             │                       ↓
             │      initialize iterators for required inputs
             │      (begin with config/brute force, then RL)
             │                       │
             │                       ↓
             └─────── are we at the end of our iterators?
                YES                  │
                                     ↓
INPUT (index, chars)     enter data into form inputs 🠤───────┐
                                     │                       │
                                     ↓                       │
SUBMIT (index)          submit form and load next page       │
                                     │                       │
                                     ↓                       │
                     ┌──────🠦 scrape the page                │
                     │               │                       │
                     │               ↓                       │
GET_LINKS            │     look for a next button ───────────┘
                     │         (classifier)        NOT FOUND
                     │               │
                     │               │ YES
                     │               ↓
SELECT_LINK (index)  └─── click the next button & load page
"""



# -*- coding: UTF-8 -*-
import lxml.html
# use threads so we can run autoscrape inside celery
from multiprocessing.pool import ThreadPool

import requests
from urllib.parse import urlparse, ParseResult

from autoscrape.backends.base.dom import DomBase


def download_stylesheet(css_url):
    response = requests.get(css_url)
    data = response.text
    if type(data) == bytes:
        return data.decode("utf-8")
    return data


class Dom(DomBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dom = self._get_dom()

    def _get_dom(self):
        dom = None
        try:
            dom = lxml.html.fromstring(self.current_html)
        # this handles trying to load XML, RSS feed, etc
        except ValueError as e:
            if "Please use bytes input" in str(e):
                html_b = bytes(self.current_html, encoding="utf-8")
                dom = lxml.html.fromstring(html_b)
            else:
                raise e
        # if our page's HTML is just an element, like an
        # iframe, without a body or html then lxml will
        # return an element surrounded by a body and html.
        # so here we make element the root and use that as
        # our base DOM.
        while True:
            parent = dom.getparent()
            if parent is None:
                break
            dom = parent
        return dom

    def element_attr(self, element, name, default=None):
        if not element.attrib:
            return default
        return element.attrib.get(name, default)

    def element_by_tag(self, tag):
        elements = self.dom.cssselect(tag)
        if not elements:
            return None
        return elements[0]

    def elements_by_path(self, xpath, from_element=None):
        if from_element is None:
            return self.dom.xpath(xpath)
        return from_element.xpath(xpath)

    def get_stylesheet(self, fetch_css=False):
        stylesheet_urls = []
        for link in self.dom.xpath("//link"):
            if not link.attrib:
                continue
            l_type = link.attrib.get("type")
            l_rel = link.attrib.get("rel")
            l_href = link.attrib.get("href")
            if l_type != "text/css" and l_rel != "stylesheet":
                continue
            css_url = self._normalize_url(l_href)
            stylesheet_urls.append(css_url)

        pool = None
        try:
            pool = ThreadPool(8)
        except OSError:
            # operating system/container doesn't support threading
            pass

        if pool is not None:
            results = pool.map(download_stylesheet, stylesheet_urls)
            pool.close()
        # fallback to single threaded in case of threading not permitted
        else:
            results = []
            for css_url in stylesheet_urls:
                results.append(download_stylesheet(css_url))

        css = "\n".join(results)
        for style in self.dom.xpath("style"):
            css += style.text_content()
        return css

    def _normalize_url(self, url):
        argnames = ['scheme', 'netloc', 'path', 'params', 'query', 'fragment']
        inheritable = ['scheme', 'netloc', 'path']
        parsed_current_url = urlparse(self.current_url)
        parsed_url = urlparse(url)

        args = []
        for argname in argnames:
            value = getattr(parsed_url, argname, None)
            if not value and argname in inheritable:
                value = getattr(parsed_current_url, argname, '')
            args.append(value)

        pr = ParseResult(*args)
        normalized = pr.geturl()
        return normalized

    def element_text(self, element, block=False):
        if block and element is not None:
            return element.text_content()
        if element is None:
            return ''
        text = element.text
        if not text:
            return ''
        return text

    def element_name(self, element):
        return element.name

    def element_tag_name(self, element):
        if element is None:
            return ""
        return element.tag



def read_file(filename):
    with open(filename) as f:
        return f.read()


def test_good_html_parse():
    html = read_file("tests/tag_test_data_page.html")
    assert(html)
    assert(len(html) > 0)



import logging
from multiprocessing import Pool
import os
import pickle

try:
    import plyvel
    import warcio
except ModuleNotFoundError:
    pass


logger = logging.getLogger('AUTOSCRAPE')


def _warc_record_sane(record):
    if record.rec_type != "response":
        return False
    if not record.rec_headers.get_header('WARC-Target-URI'):
        return False
    return True


def _warc_records(filename):
    records = []
    try:
        with open(filename, "rb") as f:
            for record in warcio.ArchiveIterator(f):
                if not _warc_record_sane(record):
                    continue
                parsed_rec = {
                    "uri": record.rec_headers.get_header('WARC-Target-URI'),
                    "payload": record.content_stream().read().strip(),
                    "headers": record.http_headers.headers,
                }
                yield parsed_rec
    except Exception as e:
        logger.error("[!] Error opening WARC file %s" % (filename))
        logger.error(e)
    return records


def _process_warcfile(filepath, filter_domain):
    found = 0
    if not filepath.endswith(".warc.gz"):
        return []
    logger.debug(" - Parsing %s" % (filepath))
    record_number = -1
    results = []
    for record in _warc_records(filepath):
        record_number += 1
        uri = record["uri"]
        if filter_domain and filter_domain not in uri:
            continue
        logger.debug("URI: %s" % (uri))
        found += 1
        uri_bytes = bytes(uri, "utf-8")
        value = pickle.dumps((filepath, record_number))
        results.append((uri_bytes, value))
    if found:
        logger.debug(" - Found %s records" % (found))
    return results


def build_warc_index(db=None, warc_directory=None, filter_domain=None):
    """
    Read through all WARC files in warc_directory and build
    an index: URL => filename, record_number
    """
    blank = True
    for rec in db.iterator():
        blank = False
        break
    if not blank:
        logger.debug("[.] Already loaded WARC index.")
        return
    logger.info("[.] Building WARC index. This might take a while...")
    _, _, filenames = list(os.walk(warc_directory))[0]
    filepaths = [(os.path.join(warc_directory, n), filter_domain) for n in filenames]
    print(filepaths[0])

    with Pool(4) as f:
        results_groups = f.starmap(_process_warcfile, filepaths)
    for results in results_groups:
        for uri_bytes, value in results:
            db.put(uri_bytes, value)



# -*- coding: UTF-8 -*-
import logging

from . import BaseScraper
from ..control import Controller
from autoscrape.backends.requests.browser import RequestsBrowser
from autoscrape.backends.selenium.browser import SeleniumBrowser


logger = logging.getLogger('AUTOSCRAPE')


class NullScraper(BaseScraper):
    """
    A test scraper that just provides direct access to scraper and
    controller. For vectorizing documents.
    """

    def __init__(self, *args, html_embeddings=None, word_embeddings=None,
                 loglevel="INFO", scraper=True, controller=False, backend=None,
                 driver="Chrome", vectorizer=False, **kwargs):
        super(NullScraper, self).setup_logging(loglevel=loglevel)

        Browser = None
        if backend == "selenium":
            Browser = SeleniumBrowser
        elif backend == "requests":
            Browser = RequestsBrowser
        else:
            raise NotImplementedError(
                "No backend found: %s" % (backend)
            )

        if scraper:
            self.scraper = Browser(driver=driver)
        if controller:
            self.control = Controller(
                html_embeddings_file=html_embeddings,
                word_embeddings_file=word_embeddings,
            )
        if vectorizer:
            self.vectorizer = self.control.vectorizer



# -*- coding: UTF-8 -*-
import logging


logger = logging.getLogger('AUTOSCRAPE')


class TextVectorizer:
    """
    # for ix, text in link_vectors
    link_vectors = self.control.vectorizer.link_vectors()
    # for ix, text in button_vectors
    button_data = self.control.vectorizer.button_vectors()
    form_vectors = self.control.vectorizer.form_vectors()
    """
    def __init__(self, scraper=None, controller=None):
        self.scraper = scraper
        self.controller = controller

    def page_vector(self, html):
        """
        Get feature vector from currently loaded page. This should
        be used to determine what type of page we're on and what action
        we ought to take (continue crawl, enter input, scrape structured
        data, etc).
        """
        return self.scraper.element_text(None, block=True)

    def form_vectors(self):
        """
        Get a feature vector representing the forms on a page. This ought
        to be used in cases where the model indicates the page may be a
        search page, but where there are multiple forms. Or where you
        just want to determine if a form is interactive data search.
        Another alternative strategy would be to try the search and then
        look at the next page.
        """
        logger.debug("[.] Loading form vectors")
        form_data = []
        for tag in self.controller.forms:
            form = self.scraper.element_by_tag(tag)
            txt = self.scraper.element_text(form, block=True)
            form_data.append(txt)
        return form_data

    def button_vectors(self):
        logger.debug("[.] Building button vectors")
        buttons_data = []
        for tag in self.controller.scraper.get_buttons():
            elem = self.scraper.element_by_tag(tag)
            value = ""
            if elem is not None:
                value = self.scraper.element_value(elem)
            text = []
            if value:
                text.append(value)
            if elem is not None:
                text.append(self.scraper.element_text(elem))
            logger.debug(" - button value: %s, text: %s" % (value, text))
            buttons_data.append(" ".join(text))
        return buttons_data

    def link_vectors(self):
        """
        Get a matrix of link vectors. These describe the text of the link
        in a way that a ML algorithm could decide how to prioritize the
        search pattern.
        """
        logger.debug("[.] Building link vectors")
        buttons_data = []
        for t in self.controller.scraper.get_clickable():
            elem = self.scraper.element_by_tag(t)
            tag_name = self.scraper.element_tag_name(elem)
            text = ""
            if elem is None:
                logger.warn("[!] Link element couldn't be found: %s" % t)
            elif tag_name != "input":
                text = self.scraper.element_text(elem).replace("\n", " ")
            elif tag_name == "input":
                value = self.scraper.element_attr(elem, "value")
                text = value.replace("\n", " ")
            buttons_data.append(text)
        return buttons_data



# -*- coding: UTF-8 -*-
import os
import re

from celery import Celery

from .scrapers.manual import ManualControlScraper


backend = "rpc://"
if os.environ.get("AUTOSCRAPE_DB_HOST"):
    backend = 'db+postgresql://%s:%s@%s/autoscrape' % (
        os.environ["AUTOSCRAPE_DB_USER"],
        os.environ["AUTOSCRAPE_DB_PASSWORD"],
        os.environ["AUTOSCRAPE_DB_HOST"]
    )


app = Celery(
    'tasks',
    broker=os.environ.get("AUTOSCRAPE_RABBITMQ_HOST"),
    backend=backend,
)

app.conf.update(
    # CELERYD_MAX_TASKS_PER_CHILD=1,
    # CELERYD_PREFETCH_MULTIPLIER=1,
    # CELERY_ACKS_LATE=True,
    # CELERY_RESULT_PERSISTENT=True,
    # CELERY_TASK_PUBLISH_RETRY=False,
    # CELERY_TASK_RESULT_EXPIRES=None,
    CELERY_TRACK_STARTED=True,
    CELERY_BROKER_HEARTBEAT=10
)


@app.task(bind=True)
def start(self, baseurl, args):
    print("Starting ManualControlScraper", baseurl, args)
    # append task ID to receiver URI
    output = args.get("output")
    if output and re.match("^https?://", output):
        if output[-1] != "/":
            output += "/"
        output += str(self.request.id)
        args["output"] = output
    scraper = ManualControlScraper(baseurl, **args)
    scraper.run()



# -*- coding: UTF-8 -*-
TEXT_EXTENSIONS = (
    # ASP Classic
    "asp",

    # ASP.NET
    "aspx",
    "axd",
    "asx",
    "asmx",
    "ashx",

    # CSS
    "css",

    # Coldfusion
    "cfm",

    # Erlang
    "yaws",

    # # Flash, treat as data
    # "swf",

    # HTML
    "html",
    "htm",
    "xhtml",
    "jhtml",
    "hta",

    # Java
    "jsp",
    "jspx",
    "wss",
    "do",
    "action",

    # JavaScript
    "js",

    # Perl
    "pl",

    # PHP
    "php",
    "php4",
    "php3",
    "phtml",

    # Python
    "py",

    # Ruby
    "rb",
    "rhtml",

    # SSI
    "shtml",

    # Other (C, perl etc.)
    # NOTE: these could also be data
    "cgi",
    "dll",

    # ASP.NET
    "axd",
    "asx",
    "asmx",
    "ashx",
    "aspx",

    # XML/Feed
    "xml",
    "rss",
    "atom",
    "svg",
)



# -*- coding: UTF-8 -*-
from collections import deque

import networkx as nx


class BFS(object):
    def __init__(self, root_node):
        # queue of remaining BFS nodes
        self.queue = deque()
        # store BFS scrape graph
        self.graph = nx.DiGraph()
        # current node (css tag)
        self.current = None

    def next(self):
        """
        Get the next CSS path node to interact with.
        """
        newcurrent = self.queue.popleft()
        self.current = newcurrent
        return newcurrent

    def add_root_node(self, node, **kwargs):
        """
        Add an isolated node, make it the current node. This
        is for creating new subgraphs inside our graph. Does
        not add any edges.
        """
        self.graph.add_node(node, **kwargs)
        self.current = node

    def add_node(self, node, **kwargs):
        """
        Add a single node to the tree, with edges connecting
        to the current node.
        """
        self.queue.append(node)
        self.graph.add_node(node, **kwargs)
        self.graph.add_edge(self.current, node)

    def add_nodes(self, nodes):
        """
        Add a list of nodes to the current node in the graph.
        This handles adding the nodes and the edges.
        """
        self.queue.extend(nodes)
        for node, meta in nodes:
            self.add_node(node, **meta)



# -*- coding: UTF-8 -*-
import re


class InputParser:
    def __init__(self, input):
        self.input = input

    def str2bool(self, string):
        if not string:
            return False
        if string.lower() in ["false", "no", "n", "0"]:
            return False
        return True

    def generate(self):
        """
        Make a form input generator by parsing our input string. Output
        is a multidimensional array, where the first dimension is
        independent searches to attempt and the second dimension is
        which inputs for fill. Example:

        From this input string:

          "i:0:a%,c:1:True,s:2:France"

        We get the following output generator:

            [
              [
                { "index": 0, "string": "a%", "type": "input" }
                { "index": 1, "action": True, "type": "checkbox" }
                { "index": 2, "string": "France", "type": "select" }
              ],
              ...,
              [
                { "index": 0, "string": "z%", "type": "input" },
                { "index": 1, "action": True, "type": "checkbox" }
                { "index": 2, "string": "France", "type": "select" }
              ]
            ]

        This will do all searches with input 0 filled with a-z and the
        remaining specified inputs filled as desired (input 1, a checkbox,
        checked and input 2, a choice select, selected to the "France"
        option).
        """
        # TODO: Find an overall way to support ranges without resulting
        # to clunky regex (which don't support ordering). Right now just
        # use GNU Parallel or something to do ranges.
        # split the independent searches first
        inputs = re.split(r'(?<!\\);', self.input)
        for inp in inputs:
            indiv_search = []
            # split the inputs to be filled per search
            indiv_inputs_list = re.split(r'(?<!\\),', inp)
            for indiv_inputs in indiv_inputs_list:
                # input type, input index, action
                itype, ix, action = indiv_inputs.split(":", 2)
                action = action.replace("\,", ",").replace("\;", ";")
                ix = int(ix)
                if itype == "i":
                    indiv_search.append({
                        "index": ix,
                        "type": "input",
                        "string": action,
                    })
                elif itype == "s":
                    indiv_search.append({
                        "index": ix,
                        "type": "select",
                        "string": action,
                    })
                elif itype == "c":
                    indiv_search.append({
                        "index": ix,
                        "type": "checkbox",
                        "action": self.str2bool(action),
                    })
                elif itype == "d":
                    # make sure our date field is correctly
                    # formatted. non-matching date fields
                    # will not make it to the input in webdriver
                    # datefmt = re.match("[0-9]{4}\-[0-9]{2}\-[0-9]{2}", action)
                    # assert datefmt, "Bad Date! Dates need to be: YYYY-MM-DD"
                    indiv_search.append({
                        "index": ix,
                        "type": "date",
                        "string": action,
                    })
                elif itype == "r":
                    indiv_search.append({
                        "index": ix,
                        "type": "radio",
                        "string": int(action),
                    })
                else:
                    raise Exception("Invalid input type found: %s" % itype)

            yield indiv_search



# -*- coding: UTF-8 -*-
import time
import logging

from autoscrape.backends.requests.browser import RequestsBrowser
from autoscrape.vectorization.text import TextVectorizer

# backends/vectorizers with optional dependencies
try:
    from autoscrape.vectorization.ebmeddings import EmbeddingsVectorizer
except ModuleNotFoundError:
    pass

try:
    from autoscrape.backends.selenium.browser import SeleniumBrowser
except ModuleNotFoundError:
    pass

try:
    from autoscrape.backends.warc.browser import WARCBrowser
except ModuleNotFoundError:
    pass


logger = logging.getLogger('AUTOSCRAPE')


class Controller:
    """
    High-level control for scraping a web page. This allows us to control
    all of the possible scraper commands in an automated way, using a set
    of indices instead of tags. This way we can present vectors of options
    to a ML model. This abstraction also returns feature matrices for pages
    and elements on the webpage.
    """

    def __init__(self, leave_host=False, driver="Firefox", browser_binary=None,
                 remote_hub="http://localhost:4444/wd/hub", output=None,
                 form_submit_natural_click=False, form_submit_wait=5,
                 warc_index_file=None, warc_directory=None,
                 force_page_wait=None, form_submit_button_selector=None,
                 load_images=False, show_browser=False, page_timeout=None,
                 html_embeddings_file=None, word_embeddings_file=None,
                 backend="selenium", vectorizer="text"):
        """
        Set up our WebDriver and misc utilities.
        """
        Browser = None
        self.backend = backend
        if backend == "selenium":
            Browser = SeleniumBrowser
        elif backend == "requests":
            Browser = RequestsBrowser
        elif backend == "warc":
            Browser = WARCBrowser
        else:
            raise NotImplementedError(
                "No backend found: %s" % (backend)
            )

        self.scraper = Browser(
            leave_host=leave_host, driver=driver,
            browser_binary=browser_binary, remote_hub=remote_hub,
            form_submit_natural_click=form_submit_natural_click,
            form_submit_wait=form_submit_wait,
            form_submit_button_selector=form_submit_button_selector,
            warc_index_file=warc_index_file, warc_directory=warc_directory,
            load_images=load_images, show_browser=show_browser,
            output=output, page_timeout=page_timeout,
        )

        Vectorizer = None
        if vectorizer == "text":
            self.vectorizer = TextVectorizer(
                scraper=self.scraper, controller=self
            )
        elif vectorizer == "embeddings":
            self.vectorizer = EmbeddingsVectorizer(
                scraper=self.scraper, controller=self,
                html_embeddings_file=html_embeddings_file,
                word_embeddings_file=word_embeddings_file,
            )
        else:
            raise NotImplementedError(
                "No vectorizer found: %s" % (vectorizer)
            )

        # this flag marks vectors as stale. when this is true and
        # we try to access the link vectors, we'll re-load them
        self.stale = True

        self.clickable = None

        # simply a list of form tags, each forms input contents is
        # contained in the self.inputs multi-dimensional array, below
        self.forms = []

        # this expands into the following format:
        # [ form_tag:
        #   [
        #     [text input tags...],
        #     [select input tags...],
        #     [checkbox input tags...]
        #   ],
        #   other forms ...,
        # ]
        self.inputs = []

        # TODO: the point of this wait is to ensure the DOM has stopped
        # mutating (loading results, etc). a proper fix for this is to
        # look at the count of DOM objects being queried for each index
        # type and detect when it stops changing.
        self.force_page_wait = force_page_wait
        if self.force_page_wait is not None:
            self.force_page_wait = int(self.force_page_wait)

    def load_indices(self):
        logger.debug("[.] Loading page vectors...")
        if self.backend == "selenium" and self.force_page_wait:
            logger.debug(" - Force waiting for %s seconds" % (
                self.force_page_wait
            ))
            time.sleep(self.force_page_wait)

        self.clickable = None
        # self.clickable = self.scraper.get_clickable()
        logger.debug(" - Getting forms")
        forms_dict = self.scraper.get_forms()
        self.forms = list(forms_dict.keys())
        logger.debug(" - Getting inputs")
        self.inputs = [tags for tags in forms_dict.values()]
        self.buttons = None # self.scraper.get_buttons()

        # logger.debug("Clickable links: %s" % (len(self.clickable)))
        # for i in range(len(self.clickable)):
        #     t = self.clickable[i]
        #     elem = self.scraper.element_by_tag(t)
        #     text = ""
        #     if elem:
        #         text = elem.text.replace("\n", " ")
        #     logger.debug("  %s - ...%s, %s" % (i, t[-25:], text))

        # logger.debug("Forms: %s:" % (len(self.forms)))
        # for i in range(len(self.forms)):
        #     t = self.forms[i]
        #     text = ""
        #     elem = self.scraper.element_by_tag(t)
        #     if elem:
        #         text = elem.text.replace("\n", " ")
        #     logger.debug("  %s - ...%s, %s" % (i, t[-25:], text))

        # logger.debug("Inputs: %s" % (len(self.inputs)))
        # for i in range(len(self.inputs)):
        #     input_group = self.inputs[i]
        #     for itype_ix in range(len(input_group)):
        #         for t in input_group[itype_ix]:
        #             elem = self.scraper.element_by_tag(t)
        #             text = ""
        #             placeholder = ""
        #             if elem:
        #                 text = elem.text.replace("\n", " ")
        #                 placeholder = elem.get_attribute("placeholder")
        #             logger.debug("  %s - ...%s, %s, %s" % (
        #                 i, t[-25:], text, placeholder))

        # logger.debug("Buttons: %s" % (len(self.buttons)))
        # for i in range(len(self.buttons)):
        #     t = self.buttons[i]
        #     elem = self.scraper.element_by_tag(t)
        #     text = ""
        #     value = ""
        #     if elem:
        #         text = elem.text.replace("\n", " ")
        #         value = elem.get_attribute("value")
        #     logger.debug("  %s - ...%s, %s, %s" % (i, t[-25:], text, value))

    def initialize(self, url):
        """
        Instantiate a web scraper, given a starting point URL. Also
        gets the links for the current page and sets its tag array.
        """
        self.scraper.fetch(url, initial=True)
        self.load_indices()

    def select_link(self, index, iterating_form=False):
        if self.clickable is None:
            logger.debug(" - Getting links")
            self.clickable = self.scraper.get_clickable()
        if index >= len(self.clickable):
            logger.error(
                "[!] Critical error: link index exceeds clickable length."
            )
            return False
        tag = self.clickable[index]
        clicked = self.scraper.click(tag, iterating_form=iterating_form)
        if clicked:
            self.load_indices()
        return clicked

    def select_button(self, index, iterating_form=False):
        if self.buttons is None:
            logger.debug(" - Getting buttons")
            self.buttons = self.scraper.get_buttons()
        tag = self.buttons[index]
        clicked = self.scraper.click(tag, iterating_form=iterating_form)
        if clicked:
            self.load_indices()
        return clicked

    def input(self, form_ix, index, chars):
        """
        Add some string to a text input under a given form.
        """
        tag = self.inputs[form_ix][0][index]
        self.scraper.input(tag, chars)

    def input_select_option(self, form_ix, index, option_str):
        """
        Select an option for a select input under a given form.
        """
        tag = self.inputs[form_ix][1][index]
        self.scraper.input_select_option(tag, option_str)

    def input_checkbox(self, form_ix, index, to_check):
        """
        Check/uncheck a checkbox input under a given form.
        """
        tag = self.inputs[form_ix][2][index]
        self.scraper.input_checkbox(tag, to_check)

    def input_date(self, form_ix, index, chars):
        """
        Select a date from an input type="date". String needs to
        be in the MM-DD-YYYY format.
        """
        tag = self.inputs[form_ix][3][index]
        self.scraper.input(tag, chars)

    def input_radio_option(self, form_ix, index, radio_index):
        """
        Select a radio checkbox from a given form, checkbox group
        index and desired value by text.
        """
        grp_tags = self.inputs[form_ix][4][index]
        tag = grp_tags[radio_index]
        self.scraper.input_checkbox(tag, True, radio=True)

    def submit(self, index):
        tag = self.forms[index]
        self.scraper.submit(tag)
        self.load_indices()

    def back(self):
        self.scraper.back()
        self.load_indices()



import unittest
import urllib

import autoscrape


class TestTagGeneration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.filename = "test_page_large.cleaned.html"
        self.url = "http://localhost:8000/%s" % (self.filename)

    def test_test_server_running(self):
        """
        Make sure our test HTML server is online. We need this to be
        loaded to continue the remaining tests.
        """
        html = None
        try:
            html = urllib.request.urlopen(self.url).read()
        except urllib.error.URLError:
            pass
        msg = "Test server not running! HINT: python -m http.server --directory ./tests/data/"
        self.assertIsNotNone(
            html, msg=msg
        )

    def _get_page_source(self):
        with open("tests/data/%s" % (self.filename), "r") as f:
            return f.read()

    def test_requests_backend_can_load_page(self):
        """
        Load the test page and make sure it matches our test page.
        """
        self.requests_browser = autoscrape.backends.requests.browser.RequestsBrowser()
        self.requests_browser.fetch(self.url)
        loaded_html = self.requests_browser.page_html
        self.assertIsNotNone(loaded_html)
        raw_html = self._get_page_source()
        self.assertEqual(raw_html, loaded_html)

    def test_selenium_backend_can_load_page(self):
        """
        Test that we can load the test page and that it doesn't get
        mutated by the browser (it's clean so this shouldn't happen).
        We do this so that later we can measure differences in the CSS
        path algorithms, not differences between represented pages.
        """
        self.selenium_browser = autoscrape.backends.selenium.browser.SeleniumBrowser()
        self.selenium_browser.fetch(self.url)
        loaded_html = self.selenium_browser.page_html
        self.assertIsNotNone(loaded_html)
        raw_html = self._get_page_source()
        self.assertEqual(raw_html, loaded_html)

    def test_backend_tags_match(self):
        self.requests_browser = autoscrape.backends.requests.browser.RequestsBrowser()
        self.requests_browser.fetch(self.url)
        self.selenium_browser = autoscrape.backends.selenium.browser.SeleniumBrowser()
        self.selenium_browser.fetch(self.url)
        sel_clickable = self.selenium_browser.get_clickable()
        req_clickable = self.requests_browser.get_clickable()
        self.assertTrue(len(sel_clickable) > 0)
        self.assertTrue(len(req_clickable) > 0)
        # self.assertEqual(len(sel_clickable), len(req_clickable))
        # self.assertEqual(sel_clickable, req_clickable)


if __name__ == "__main__":
    unittest.main()



# -*- coding: UTF-8 -*-
import logging

from . import BaseScraper

from autoscrape.backends.requests.browser import RequestsBrowser


logger = logging.getLogger('AUTOSCRAPE')


class TestScraper(BaseScraper):
    def __init__(self, baseurl, maxdepth=10, loglevel=None):
        """
        Initialize our scraper and get the first page.
        """
        super(TestScraper, self).setup_logging(loglevel=loglevel)
        self.scraper = RequestsBrowser()
        self.scraper.fetch(baseurl)
        self.maxdepth = maxdepth

    def run(self, depth=0, tags=None):
        """
        This is the main recursive depth-first search of a site. It
        doesn't do anything but crawl a site DFS and ensure the tagging
        and web engine is working as it should.
        """
        if depth > self.maxdepth:
            logger.debug("Maximum depth %s reached, returning..." % depth)
            self.scraper.back()
            return

        logger.debug("** DEPTH %s" % depth)

        if not tags:
            tags = self.scraper.get_clickable()

        logger.debug(
            "All tags at this depth \n    %s" % ("\n    ").join(tags))

        for tag in tags:
            logger.debug("Attempting click on tag \n    %s" % tag)

            if self.scraper.click(tag):
                logger.debug("Clicked! Recursing ...")
                self.run(
                    depth=depth + 1, tags=self.scraper.get_clickable())

        logger.debug("Going back...")
        self.scraper.back()



# -*- coding: UTF-8 -*-
import logging
import re

import numpy as np


logger = logging.getLogger('AUTOSCRAPE')


class Embedding:
    def __init__(self, embeddings=None, t2id=None, id2t=None):
        self.embeddings = embeddings
        self.t2id = t2id
        self.id2t = id2t
        self.N, self.dim = embeddings.shape


class EmbeddingsVectorizer:
    def __init__(self, html_embeddings_file=None, word_embeddings_file=None,
                 scraper=None, controller=None, loglevel=None):
        """
        Initialize our vectorizer with paths to the relevant word
        embedding files for our vectorization routines.

        Representations:
            - Element text vector (fasttext)
            - Element path vector ( body -> form -> div ... -> button )

        Given state, q-function gets action-state pairs. We take
        the action giving the highest q-value.
        """
        self.scraper = scraper
        self.controller = controller

        self.html = None
        if html_embeddings_file:
            logger.debug("[.] Loading HTML embeddings")
            self.html = self.load_embedding(html_embeddings_file)

        self.word = None
        if word_embeddings_file:
            logger.debug("[.] Loading word embeddings")
            self.word = self.load_embedding(word_embeddings_file)

    def embeddings_length(self, path):
        N = 0
        with open(path, "r") as f:
            for line in f:
                if N == 0 and re.match("^[0-9]+\s[0-9]+$", line):
                    continue
                N += 1
            key, data = line.split(' ', 1)
            vec = [float(d) for d in data.split()]
            dim = len(vec)
        return N, dim

    def load_embedding(self, path):
        logger.info("[+] Loading embedding file %s..." % path)
        N, dim = self.embeddings_length(path)
        logger.info(" - vocab size: %s, dim: %s" % (
            N, dim
        ))
        logger.debug(" - Allocating embedding matrix...")
        # token to ID (embedding row)
        t2id = dict()
        # ID to token
        id2t = dict()
        # embedding matrix
        embeddings = np.zeros(shape=(N, dim))
        logger.debug(" - Reading embeddings into memory...")
        outputs = [(N // 10) * i for i in range(10)]
        with open(path, "r") as f:
            embed_id = 0
            for line in f:
                if embed_id == 0 and re.match("^[0-9]+\s[0-9]+$", line):
                    continue
                if embed_id in outputs:
                    pct_done = (embed_id / float(N)) * 100
                    logger.info(" - %0.4f%% complete" % (pct_done))
                key, data = line.split(' ', 1)
                vec = [float(d) for d in data.split()]
                embeddings[embed_id, :] = vec
                t2id[key] = embed_id
                id2t[embed_id] = key
                embed_id += 1

        logger.debug(" - Embeddings matrix: %s x %s" % embeddings.shape)
        return Embedding(
            embeddings=embeddings,
            t2id=t2id,
            id2t=id2t,
        )

    def html_to_vector(self, html):
        x = np.zeros(self.html.dim)
        N = 0.0
        for t in html:
            N += 1
            if re.match("\s", t):
                t = "</s>"
            id = self.html.t2id[t]
            x += self.html.embeddings[id]
        return x / N

    def text_to_vector(self, text):
        x = np.zeros(self.word.dim)
        N = 0.0
        for t in re.split("[^A-Za-z]", text):
            t = t.strip().lower()
            if not t:
                continue
            N += 1
            if re.match("\s", t):
                t = "</s>"
            try:
                id = self.word.t2id[t]
            except Exception as e:
                logger.warn("Skipping word=%s,  Error=%s" % (
                    t, e
                ))
                continue
            x += self.word.embeddings[id]
        return x / N

    def element_to_position_vector(self, element):
        return np.array([0.0])

    def vectorize(self, html, text, element=None):
        x_html = self.html_to_vector(html)
        x_text = self.text_to_vector(text)
        concat_array = [x_html, x_text]
        if element:
            x_pos = self.element_to_position_vector(element)
            concat_array.append(x_pos)
        x = np.concatenate(concat_array)
        return x

    def page_vector(self):
        raise NotImplementedError(
            "EmbeddingsVectorizer.page_vector not implemented"
        )

    def form_vectors(self):
        raise NotImplementedError(
            "EmbeddingsVectorizer.form_vectors not implemented"
        )

    def button_vectors(self):
        raise NotImplementedError(
            "EmbeddingsVectorizer.button_vectors not implemented"
        )

    def link_vectors(self):
        raise NotImplementedError(
            "EmbeddingsVectorizer.link_vectors not implemented"
        )



#!/usr/bin/env python
from os import path
import setuptools


def get_long_description():
    BASEDIR = path.abspath(path.dirname(__file__))
    with open(path.join(BASEDIR, 'README.rst'), encoding='utf-8') as f:
        return f.read()


setuptools.setup(
    name='autoscrape',
    version='1.6.13',
    description='An automated, programming-free web scraper for interactive sites',
    long_description=get_long_description(),
    author='Brandon Roberts',
    author_email='brandon@bxroberts.org',
    url='https://github.com/brandonrobertz/autoscrape-py',
    license='AGPLv3',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities'
    ],
    packages=[
        'autoscrape',
        'autoscrape.cli',
        'autoscrape.util',
        'autoscrape.backends',
        'autoscrape.backends.base',
        'autoscrape.backends.selenium',
        'autoscrape.backends.requests',
        'autoscrape.backends.warc',
        'autoscrape.scrapers',
        'autoscrape.search',
        'autoscrape.vectorization',
    ],
    entry_points={
        'console_scripts': [
            'autoscrape = autoscrape.cli.scrape:main',
        ]
    },
    install_requires=[
        'lxml>=4.3.0',
        'html5lib>=1.0.1',
        'docopt>=0.6.2',
        'cssselect>=1.1.0',
        'requests>=2.22.0',
    ],
    extras_require={
        'selenium-backend': [
            'selenium>=3.141.0,<4.0.0',
        ],
        'warc-backend': [
            'warcio>=1.7.3,<2.0.0',
            'plyvel==1.2.0',
        ],
        'embeddings-vectorizer': [
            'numpy>=1.15.0'
        ],

        'graph': [
            'networkx>=2.2',
        ],
        'all': [
            'selenium>=3.141.0,<4.0.0',
            'networkx>=2.2',
            'numpy>=1.15.0',
            'warcio>=1.7.3,<2.0.0',
            'plyvel==1.2.0',
        ],
    }
)



# -*- coding: utf-8 -*-
from autoscrape.backends.base.tags import TaggerBase


class BrowserBase(TaggerBase):
    def _no_tags(self, data, l_type="path"):
        clean = []
        if type(data) == tuple:
            data = list(data)
        for p in data:
            name, t_args, kwargs = p
            args = list(t_args)
            if name == "click":
                if not args:
                    continue
                args[0] = "[tag]"
            clean.append((name, args, kwargs))
        return clean

    def click(self, tag, **kwargs):
        self.path.append((
            "click", [tag], {"url": url}
        ))
        node = "Click\n text: %s\n hash: %s" % (text, hash)
        node_meta = {
            "click": tag,
            "click_text": text,
            "click_iterating_form": None,
        }
        self.graph.add_node(
            node,
            **node_meta
        )
        self.graph.move_to_node(node)

    def fetch(self, url, initial=False):
        self.graph.add_root_node(node, url=url, action="fetch")

    def back(self):
        self.graph.move_to_parent()

    def input(self, tag, input):
        self.path.append(("input", ("", input,), {}))
        action = {
            "action": "input",
            "text": input,
            "tag": tag,
        }
        self.graph.add_action_to_current(action)

    def submit(self, tag, add_node=True):
        self.path.append(("submit", (tag,), {}))
        node = "Submit\n tag: %s" % (tag)
        node_meta = {
            "submit": tag,
        }
        self.graph.add_node(node, **node_meta)
        self.graph.move_to_node(node)

    @property
    def infinite_loop_detected(self):
        return False



# -*- coding: UTF-8 -*-
"""
AUTOSCRAPE - Interactively crawl, find searchable forms,
input data to them and scrape data on the results, from an
initial BASEURL.

Usage:
    autoscrape [options] BASEURL

General Options:
    --backend BACKEND
        The backend to use. Currently one of "selenium", "requests" or
        "warc".  The requests browser is only capable of crawling, but
        is approximately 2-3.5x faster. WARC is for emulating browsing
        through Common Crawl archival data.
        [default: selenium]

    --loglevel LEVEL
        Loglevel, note that DEBUG is extremely verbose.
        [default: INFO]

    --quiet
        This will silence all logging to console.

Crawl-Specific Options:
    --maxdepth DEPTH
        Maximum depth to crawl a site (in search of form
        if the option --form-match STRING is specified,
        see below). Setting to 0 means don't crawl at all,
        all operations are limited to the BASEURL page.
        Setting to -1 means unlimited maximum crawl depth.
        [default: 10]

    --max-pages NUM
        Maximum number of unique pages, in total, to fetch.
        AutoScrape will stop crawling once this is hit.

    --leave-host
        By default, autoscrape will not leave the host given
        in the BASEURL. This option lets the scraper leave
        the host.

    --only-links MATCH_STRING
        A whitelist of links to follow. All others will
        be ignored. Can be a string or a regex with
        multiple strings to match separated by a pipe
        (|) character.

    --ignore-links MATCH_STRING
        This option can be used to remove any links matching
        MATCH_STRING (can be a regex or just a string match)
        from consideration for clicking. Accepts the same
        argument format as --only-links.

    --link-priority SORT_STRING
        A string to sort the links by. In this case, any link
        containing "SORT_STRING" will be clicked before any other
        links. In most cases you probably want to use the
        whitelist, --only-links, option.

    --ignore-extensions IGNORE_EXTENSIONS
        Don't click on or download URLs pointing to files with
        these extensions.

    --result-page-links MATCH_STRINGS_LIST
        If specified, AutoScrape will click on any links matching
        this string when it arrives on a search result page.

Interactive Form Search Options:
    --form-match SEARCH_STRING
        The crawler will identify a form to search/scrape if it
        contains the specified string. If matched, it will be
        interactively scraped using the below instructions.

    --input INPUT_DESCRIPTION
        Interactive search descriptor. This describes how to
        interact with a matched form. The inputs are
        described in the following format:

        "c:0:True,i:0:atext,s:1:France:d:0:1991-01-20,r:1:0"

        A single-input type can be one of four types:
        checkbox ("c"), input box ("i"), option select
        ("s"), radio checkbox ("r"),  and date inputs
        ("d", with inputs in the  "YYYY-MM-DD" format).
        The type is separated by a colon, and the input
        index position is next. (Each input type has its
        own list, so a form with one input, one checkbox,
        and one option select, will all be at index 0.)
        The final command, sepearated by another colon,
        describes what to do with the input.

        Multiple inputs are separated by a comma, so you can
        interact with multiple inputs before submitting the
        form.

        To illustrate this, the above command does the following:
            - first input checkbox is checked (uncheck is False)
            - first input box gets filled with the string "first"
            - second select input gets the "France" option chosen
            - first date input gets set to Jan 20, 1991
            - the second radio checkbox group will have its
              first option selected

    --next-match NEXT_BTN_STRING
        A string to match a "next" button with, after
        searching a form.  The scraper will continue to
        click "next" buttons after a search until no matches
        are found, unless limited by the --formdepth option
        (see below).

    --formdepth DEPTH
        How deep the scraper will iterate, by clicking
        "next" buttons. Zero means infinite depth.
        [default: 0]

    --form-submit-button-selector XPATH_SELECTOR
        A XPath selector indicating a submit button/element
        to click. Use this when the submit button to a form
        doesn't have any text or is a strange element type.

    --form-submit-natural-click
        Some webpages make clicking a link element difficult
        due to JavaScript onClick events. In cases where a
        click does nothing, you can use this option to get
        the scraper to emulate a mouse click over the link's
        poition on the page, activating any higher level JS
        interactions.

    --form-submit-wait SECONDS
        How many seconds to force wait after a submit to a form.
        This should be used in cases where the builtin
        wait-for-page-load isn't working properly (JS-heavy
        pages, etc). [default: 5]

Webdriver-Specific and General Options:
    --page-timeout SECONDS
        Selenium has a lot of timeout settings used for
        determining when a page has loaded. In cases where
        pages take a very long time, this may need to be
        increased. If you have a fast-loading page, but
        AutoScrape is waiting for load, you can set this
        lower to speed things up. [default: 30]

    --force-page-wait SECONDS
       This forces AutoScrape to wait for the specified number
       of seconds after performing a navigation action.
       Increase this for slow sites or oddly loading single page
       apps, decrease this for fast sites or crawls.
       [default: 1]

    --load-images
        By default, images on a page will not be fetched.
        This speeds up scrapes on sites and lowers bandwidth
        needs. This option fetches all images on a page.

    --show-browser
        By default, we hide the browser during operation.
        This option displays a browser window, mostly
        for debugging purposes.

    --driver DRIVER
        Which browser to use. Current support for "Firefox",
        "Chrome", and "remote". [default: Firefox]

    --browser-binary PATH_TO_BROWSER
        Path to a specific browser binary. If left blank
        selenium will pull the browser found on your path.

    --remote-hub URI
        If using "remote" driver, specify the hub URI to
        connect to. Needs the proto, address, port, and path.
        [default: http://localhost:4444/wd/hub]

WARC Options:
    --warc-directory PATH_TO_WARCS
        Path to the folder containing GZipped WARC files. These can be
        downloaded from Common Crawl. Required when using the "warc"
        backend.

    --warc-index-file PATH_TO_LEVELDB
        Path to the level DB database holding the URL-to-file
        index: URL => (filename, record_number)
        This will be generated from the WARCS in the --warc-directory
        speficied if it's not already. Required when using the "warc"
        backend.

Data Saving Options:
    --output DIRECTORY_OR_URL
        If specified, this indicates where to save pages during a
        crawl. This directory will be created if it does not
        currently exist.  This directory will have several
        sub-directories that contain the different types of pages
        found (i.e., search_pages, data_pages, screenshots).
        This can also accept a URL (i.e., http://localhost:5000/files)
        and AutoScrape will POST to that endpoint with each
        file scraped.
        [default: autoscrape-data]

    --keep-filename
        By default, we hash the files in a scrape in order to
        account for dynamic content under a single-page app
        (SPA) website implmentation. This option will force
        the scraper to retain the original filename, from the
        URL when saving scrape data.

    --save-screenshots
        This option makes the scraper save screenshots of each
        page, interaction, and search. Screenshots will be
        saved to the screenshots folder of the output dir.

    --full-page-screenshots
        By default, we only save the first displayed part of the
        webpage. The remaining portion that you can only see
        by scrolling down isn't captured. Setting this option
        forces AutoScrape to scroll down and capture the entire
        web content. This can fail in certain circumstances, like
        in API output mode and should be used with care.

    --save-graph
        This option allows the scraper to build a directed graph
        of the entire scrape and will save it to the "graph"
        subdirectory under the output dir. The output file
        is a timestamped networkx pickled graph.

    --disable-style-saving
        By default, AutoScrape saves the stylesheets associated
        with a scraped page. To save storage, you can disable this
        functionality by using this option.
"""
import logging

from docopt import docopt

import autoscrape


logger = logging.getLogger('AUTOSCRAPE')


def main():
    docopt_args = docopt(__doc__)

    BASEURL = docopt_args.pop("BASEURL")

    # strip the -- and convert - to _
    args = {}
    for option in docopt_args:
        args[option[2:].replace('-', '_')] = docopt_args[option]

    # configure stdout logging
    docopt_args["stdout"] = True
    if "quiet" in args:
        quiet = args.pop("quiet")
        args["stdout"] = not quiet

    scraper = autoscrape.ManualControlScraper(BASEURL, **args)

    logger.debug("AutoScrape starting with arguments: %s" % (docopt_args))
    scraper.run()



# -*- coding: utf-8 -*-
import logging

from urllib.parse import urlparse

from autoscrape.backends.base.dom import DomBase


logger = logging.getLogger('AUTOSCRAPE')


class TaggerBase(DomBase):
    """
    Generates tags from a given page that can be used, in a stateless manner,
    to refer to unique elements on a web page.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.base_host = urlparse(self.current_url).netloc

    def tag_from_element(self, element):
        """
        For a given element, return a path (a.k.a. a "tag") leading
        to it. Path can be CSS, XPath, or whatever the backend supports.
        """
        raise NotImplementedError("Tagger.tag_from_element not implemented")

    def clickable_sanity_check(self, element, href=None):
        """
        Check a series of element nodes, checking their attributes and other
        attributes, determining if a element is actually 'clickable'. This
        check determines which nodes will end up as clickable options for
        the scraper on this page.
        """
        if not href:
            raw_href = self.element_attr(element, "href")
            if not raw_href:
                return False

            if hasattr(self, "_normalize_url"):
                href = self._normalize_url(raw_href).split("#")[0]
            else:
                href = raw_href

        if href.split("#")[0] == self.current_url:
            return False

        # skip any weird protos ... we whitelist notrmal HTTP,
        # anchor tags and blank tags (to support JavaScript & btns)
        if href and \
           not href.startswith("//") and \
           not href.startswith("https:") and \
           not href.startswith("http:") and \
           not href.startswith("javascript"):
            return False

        # Don't leave base host ... configurable?
        elem_host = urlparse(href).netloc
        if elem_host and not self.leave_host and elem_host != self.base_host:
            return False

        return True

    def get_clickable(self, path="//a"):
        """
        Get all clickable element tags on the current page.
        """
        tags = []
        a_elems = self.elements_by_path(path)
        for element in a_elems:
            if not self.clickable_sanity_check(element):
                continue

            tag = self.tag_from_element(element)
            # No way to get back to here, so we can't use it
            if not tag:
                logger.warn("No tag for element %s" % (element))
                return False

            tags.append(tag)
        return tags

    def get_inputs(self, form=None, itype=None, root_node=None):
        """
        Get inputs, either for full page or by a form WebElement.
        Returns a list of tags. itype can be one of "text", "select",
        "checkbox", or None (all types), indicating the type of input.
        """
        x_path = "//input"
        if itype == "select":
            x_path = "//select"
        elif itype:
            x_path = "//input[@type='%s']" % (itype)

        elem = root_node
        tags = []
        if form is not None:
            elem = form
            x_path = ".%s" % x_path

        # radio checkboxes are grouped by name
        if itype == "radio":
            # store the radio group position by name here
            radio_names = []
            elems = self.elements_by_path(x_path, from_element=elem)
            for radio in elems:
                name = self.element_attr(radio, "name")
                if name not in radio_names:
                    radio_names.append(name)
                    tags.append([])
                radio_group_ix = radio_names.index(name)
                radio_tag = self.tag_from_element(radio)
                if not radio_tag:
                    logger.warn("No tag for radio %s" % radio)
                    continue
                tags[radio_group_ix].append(radio_tag)

        # all the rest are flat
        else:
            elems = self.elements_by_path(x_path, from_element=elem)
            for inp in elems:
                # this should weed out hidden/non-displayed inputs ...
                if not self.element_displayed(inp):
                    continue
                # .. and just in case it didn't
                if self.element_attr(inp, "type") == "hidden":
                    continue
                input_tag = self.tag_from_element(inp)
                if not input_tag:
                    logger.warn("No tag for input %s" % input)
                    continue

                tags.append(input_tag)

        return tags

    def get_forms(self):
        """
        Get all tags to forms on a page and their respective text
        inputs. Tags are returned in a dict, with the form tag as
        the key and a list of input CSSPaths under the form.
        """
        x_path = "//form"
        forms = self.elements_by_path(x_path)

        tags = {}
        for elem in forms:
            if not self.element_displayed(elem):
                continue
            tag = self.tag_from_element(elem)
            if not tag:
                logger.warn("No tag for element %s" % elem)
                continue

            tags[tag] = [
                self.get_inputs(form=elem, itype="text"),
                self.get_inputs(form=elem, itype="select"),
                self.get_inputs(form=elem, itype="checkbox"),
                self.get_inputs(form=elem, itype="date"),
                self.get_inputs(form=elem, itype="radio"),
            ]

        return tags

    def get_buttons(self, in_form=False, path=None):
        """
        Return all tags leading to a form link, button, or submit input button,
        optionally given a base form to look from. This is used to identify
        clickable things related to forms.
        """
        x_path = path or "|".join([
            "//form//a", "//button", "//input[@type='button']",
            "//input[@type='submit']", "//table//a",
        ])
        btns = self.elements_by_path(x_path)

        tags = []
        for elem in btns:
            if not self.element_displayed(elem):
                continue
            tag = self.tag_from_element(elem)
            if not tag:
                logger.warn("No tag for element %s" % elem)
                continue
            tags.append(tag)
        return tags



# -*- coding: utf-8 -*-
import hashlib
import time
import logging
import os
import re
import sys

try:
    from selenium import webdriver
    from selenium.common.exceptions import (
        TimeoutException, StaleElementReferenceException,
        NoSuchElementException, ElementNotInteractableException,
        InvalidElementStateException, WebDriverException,
    )
    from selenium.webdriver.common.by import By
    from selenium.webdriver.common.keys import Keys
    from selenium.webdriver.firefox.firefox_binary import FirefoxBinary
    from selenium.webdriver.support.ui import WebDriverWait, Select
    from selenium.webdriver.support import expected_conditions as EC
except ModuleNotFoundError:
    # we haven't installed selenium backend deps
    pass

from autoscrape.backends.base.browser import BrowserBase
from autoscrape.backends.selenium.tags import Tagger
from autoscrape.search.graph import Graph


logger = logging.getLogger('AUTOSCRAPE')


class SeleniumBrowser(BrowserBase, Tagger):
    # override a None passed to page_timeout because we need
    # to set an integer value
    DEFAULT_TIMEOUT = 30

    def __init__(self, driver="Firefox", leave_host=False,
                 load_images=False, form_submit_natural_click=False,
                 form_submit_wait=5, output=None, show_browser=False,
                 form_submit_button_selector=None,
                 browser_binary=None, page_timeout=None,
                 remote_hub="http://localhost:4444/wd/hub", **kwargs):
        try:
            webdriver
        except NameError:
            logger.error(
                "Tried to use selenium backend but Selenium isn't"
                " installed. (Hint: pip install autoscrape[selenium-backend])"
                " Exiting."
            )
            sys.exit(1)

        # toggle the various timeouts in selenium
        self.timeout = page_timeout
        if self.timeout is None:
            self.timeout = self.DEFAULT_TIMEOUT

        # Needs geckodriver:
        # https://github.com/mozilla/geckodriver/releases
        # Version 0.20.1 is recommended as of 14/07/2018
        if driver == "Firefox":
            logger.debug(" - Starting Firefox")
            firefox_options = webdriver.firefox.options.Options()
            if not show_browser:
                logger.debug(" - Headless mode enabled")
                # override beause sometimes FF/Selenium/Geckodriver
                # will ignore the headless options
                os.environ["MOZ_HEADLESS"] = "1"
                firefox_options.add_argument("--headless")
                firefox_options.add_argument("-headless")
                firefox_options.headless = True
            firefox_profile = webdriver.FirefoxProfile()
            if not load_images:
                # disable images
                firefox_profile.set_preference(
                    'permissions.default.image', 2
                )
            #  disable flash
            firefox_profile.set_preference(
                'dom.ipc.plugins.enabled.libflashplayer.so', 'false'
            )
            firefox_profile.set_preference(
                'security.fileuri.strict_origin_policy', 'false'
            )
            binary = None
            if browser_binary is not None:
                logger.debug(" - Using binary: %s" % (browser_binary))
                binary = FirefoxBinary(browser_binary)
            self.driver = webdriver.Firefox(
                options=firefox_options,
                firefox_profile=firefox_profile,
                firefox_binary=binary,
            )
            self.driver.set_page_load_timeout(self.timeout)

        # this requires chromedriver to be on the PATH
        # if using chromium and ubuntu, apt install chromium-chromedriver
        elif driver == "Chrome":
            logger.debug(" - Starting Chrome")
            chrome_options = webdriver.ChromeOptions()
            if not show_browser:
                chrome_options.add_argument("--headless")
                chrome_options.add_argument("--disable-web-security")
                chrome_options.add_argument("--window-size=1920x1080")
            prefs = {
                "profile.managed_default_content_settings.images": 2
            }
            chrome_options.add_experimental_option("prefs", prefs)
            if browser_binary is not None:
                logger.debug(" - Using binary: %s" % (browser_binary))
                options.binary_location = browser_binary
            self.driver = webdriver.Chrome(chrome_options=chrome_options)

        elif driver == "remote":
            logger.debug(" - Starting remote browser")
            self.driver = webdriver.Remote(
                command_executor=remote_hub,
                desired_capabilities={
                    "browserName": "chrome",
                    "goog:chromeOptions": {
                        "args": [
                            "--disable-logging",
                            "--headless",
                            "--window-size=1920x1080",
                        ],
                        "prefs": {
                            "profile.managed_default_content_settings.images": 2
                        },
                        "extensions": [],
                    }
                })
        else:
            raise NotImplementedError("No driver found: %s, exiting." % (
                driver
            ))

        # set of clicked elements
        self.visited = set()
        # queue of the path that led us to the current page
        # this is in the form of (command, *args, **kwargs)
        self.path = []
        # store the values of the history stack length (via JS) for use in determining
        # whether or not we actually navigated somewhere. stored in window, hist
        # pairs: [[win_1, 1], ..., [win_N, N]]
        self.history_stack = []
        # used for detecting infinite loop inside a form (if the 'next' button
        # never does disabled once at the end of results)
        self.page_hashes = []
        # tree building
        self.graph = Graph()
        # sometimes the firefox driver loses its pipe to the browser. in
        # these cases, we can retry this number of times
        self.broken_pipe_retries = 3
        # setting to False, ensures crawl will stay on same host
        self.leave_host = leave_host
        # characters that need to be escaped if found inside an ID tag
        self.css_escapables = ".:"
        self.form_submit_natural_click = form_submit_natural_click
        self.form_submit_wait = form_submit_wait
        # selector to locate submit button, overrides all other
        # selection strategies
        self.form_submit_button_selector = form_submit_button_selector

    def _driver_exec(self, fn, *args, **kwargs):
        """
        Wrap all driver function calls with broken pipe handling. This
        is a workaround for geckodriver's problem of killing connections
        or lack of persistent connection. Geckodriver 0.20.1 doesn't seem
        to have the issue ... as it appears to be related to a very
        short keep-alive in 0.21.0.
        """
        pipe_retries = 0
        while True:
            try:
                return fn(*args, **kwargs)
            except BrokenPipeError as e:
                if pipe_retries > self.broken_pipe_retries:
                    msg = "Exceeded max broken pipe retries: %s"
                    logger.debug(msg % self.broken_pipe_retries)
                    msg = "Broken pipe err. fn: %s, args: %s, kwargs: %s"
                    logger.error(msg % (fn, args, kwargs))
                    raise e
            except TypeError as e:
                if "not callable" in str(e):
                    return self._driver_exec(lambda: fn)
                    # don't increment pipe_retries here. we just need to
                    # convert our property into a callable for the next
                    # iteration
                    continue
                raise e

            pipe_retries += 1

    def __del__(self):
        if hasattr(self, "driver") and self.driver:
            try:
                self.driver.quit()
            except Exception:
                pass

    def _get_history_depth(self):
        """
        Retrieve the browser history depth for use in detecting
        whether or not we actually navigated to a new page.
        """
        script = "return window.history.length;"
        return self._driver_exec(self.driver.execute_script, script)

    def _get_open_windows(self):
        """
        Get the handles of each open browser window and add any new
        ones to the open window stack.
        """
        return self.driver.window_handles

    def _switch_to_window(self, window_id):
        self.driver.switch_to.window(window_id)

    def _switch_to_iframe(self, frame_id):
        self.driver.switch_to.frame(frame_id)

    def _wait_check(self, driver):
        """
        This is the check that gets ran to determine whether
        the page is loaded or not.
        """
        logger.debug(" - Waiting for page to load (document.readyState)...")
        s = time.time()
        script = """
          var callback = arguments[arguments.length - 1];
          setTimeout(callback(document.readyState), %s);
          document.addEventListener('readystatechange', event => {
            console.log(event.target.readyState);
            if (event.target.readyState === 'interactive') {
            }
            else if (event.target.readyState === 'complete') {
              callback(document.readyState);
            }
          });
        """ % (5000)
        result = self._driver_exec(self.driver.execute_async_script, script)
        logger.debug(" - Page loaded in %s seconds" % (time.time() - s))
        return result == "complete"

    def _loadwait(self, fn, *args, **kwargs):
        """
        Run a driver interaction function, wait for the page to
        become ready, and handle any broken pipe errors
        """
        start = time.time()
        check_alerts = False
        if "check_alerts" in kwargs:
            check_alerts = kwargs["check_alerts"]
            del kwargs["check_alerts"]

        # get any element as a reference for staleness check
        elem = self.driver.find_element_by_xpath("//*")

        self._driver_exec(fn, *args, **kwargs)
        time.sleep(1)

        if check_alerts:
            logger.debug("[.] Checking for popup alerts...")
            try:
                WebDriverWait(self.driver, 1).until(
                    EC.alert_is_present(),
                    'Waiting for alert timed out'
                )
                alert = self.driver.switch_to_alert()
                alert.accept()
                logger.debug("[.] Alert accepted!")
            except TimeoutException:
                pass

        # wait for the page to become ready, up to 30s, checks every 0.5s
        logger.debug(" - Performing native WebDriverWait...")
        wait = WebDriverWait(self.driver, self.timeout)
        wait.until(self._wait_check)

        t = time.time() - start
        logger.debug(" - Wait for load succeeded in %s" % t)

    def scrolltoview(self, elem):
        """
        Scroll to an element before we interact with it.
        """
        script = "arguments[0].scrollIntoView();"
        try:
            self._driver_exec(self.driver.execute_script, script, elem)
        except ElementNotInteractableException as e:
            pass

    def fetch(self, url, **kwargs):
        """
        Fetch a page from a given URL (entry point, typically). Most of
        the time we just want to click a link or submit a form using
        webdriver.
        """
        logger.info("[.] Fetching %s" % url)
        self._loadwait(self.driver.get, url)
        self.path.append(("fetch", (url,), {}))

        depth = self._get_history_depth()
        window = self._get_open_windows()[-1]
        self.history_stack.append([window, depth])

        node = "Fetch\n url: %s" % url
        self.graph.add_root_node(node, url=url, action="fetch")

    def back(self):
        logger.info("[+] Going back...")
        logger.debug(" - current path-length=%s path=%s" % (
            len(self.path), self._no_tags(self.path),
        ))

        # only go 'back' if the history depth changed from
        # now to the previous (back) value. if they're the
        # same, we just clicked something inside the page
        if len(self.history_stack) > 1:
            window, depth = self.history_stack.pop()
            prev_win, prev_depth = self.history_stack[-1]
            # same window, we just incremented the history stack once
            if window == prev_win and depth != prev_depth:
                logger.debug(" - History depth changed since last action.")
                # going backward should be the only time we have a
                # frame focused, where we want to be on the main page
                self.driver.switch_to.default_content()
                self._loadwait(self.driver.back)
            # time to close an open window, focus old
            elif window != prev_win:
                logger.debug(" - Going back to previous window")
                # close current window (make sure we've focused proper one)
                self._switch_to_window(window)
                self.driver.close()
                # focus previous one
                self._switch_to_window(prev_win)
        self.path.pop()
        self.graph.move_to_parent()

    def disable_target(self, elem):
        """
        Look for targets either non-blank or not _self and set to _self.
        This needs to be a JavaScript injected script with element as param.
        """
        target = self.element_attr(elem, "target")
        if not target or target == "_self":
            return

        script = "arguments[0].target='_self';"
        self._driver_exec(self.driver.execute_script, script, elem)

    def click(self, tag, iterating_form=False):
        """
        Click an element by a given tag. Returns True if the link
        hasn't been visited and was actually clicked.
        """
        elem = self.element_by_tag(tag)
        if not elem:
            logger.warn("Element by tag not found. Tag: %s" % tag)
            return False

        try:
            wait = WebDriverWait(self.driver, self.timeout)
            wait.until(EC.element_to_be_clickable((By.CSS_SELECTOR, tag)))
        except TimeoutException as e:
            pass

        hash_parts = []
        hash_parts.append(self._driver_exec(elem.tag_name))
        href = self._driver_exec(elem.get_attribute, "href")
        hash_parts.append(href)
        hash_parts.append(self._driver_exec(elem.get_attribute, "onclick"))
        text = self._driver_exec(elem.text)
        hash = "|".join([h for h in hash_parts if h])
        logger.debug(" - Link hash : %s" % (hash))

        # only enable this for URL links
        # if href and hash in self.visited and not iterating_form:
        if hash in self.visited and not iterating_form:
            logger.debug("Link already clicked! Hash: %s" % (hash))
            return False

        self.visited.add(hash)
        self.disable_target(elem)
        self.scrolltoview(elem)

        try:
            self._loadwait(elem.click)
        except TimeoutException as e:
            # this is a fix for landing on the "view image" page
            # that browsers now all include. these cause a timeout
            # error w/ selenium. override the exception here, which
            # would tell the scraper we didn't change pages, and the
            # rest should work
            IMAGE_TYPES = ".*\.(jpg|jpeg|png|gif|bmp)"
            if re.match(IMAGE_TYPES, self.page_url):
                logger.debug("[*] We're on a image viewer page!")
            # some other problem. for now lets assume we didn't change
            # pages until we find otherwise.
            else:
                logger.error("[!] Scraper operation timed out!!")
                logger.error("[!] Current URL: %s" % (self.page_url))
                return False
        except Exception as e:
            logger.debug("[!] Error clicking: %s" % (e))
            logger.debug("[!] Current URL: %s" % (self.page_url))
            return False

        # apply form submit waits to 'next' button clicks
        if iterating_form and self.form_submit_wait:
            logger.debug(
                " - Forcing post-submit wait period of %ss" %
                self.form_submit_wait
            )
            time.sleep(self.form_submit_wait)

        depth = self._get_history_depth()
        window = self._get_open_windows()[-1]

        if len(self.history_stack) > 1:
            prev_window = self.history_stack[-1][0]
            if window != prev_window:
                logger.debug(" - Focusing new window: %s" % (window))
                self._switch_to_window(window)

        self.history_stack.append([window, depth])

        self.path.append((
            "click", (tag,), {"iterating_form": iterating_form}
        ))
        node = "Click\n text: %s\n hash: %s" % (text, hash)
        node_meta = {
            "click": tag,
            "click_text": text,
            "click_iterating_form": iterating_form,
        }
        self.graph.add_node(
            node,
            **node_meta
        )
        self.graph.move_to_node(node)
        return True

    def expand_key_substitutions(self, input):
        """
        This split an input string into a series of either
        plain text chunks of webdriver Key objects.
        """
        replacements = {
            "[:left:]": Keys.ARROW_LEFT,
            "[:right:]": Keys.ARROW_RIGHT,
            "[:down:]": Keys.ARROW_DOWN,
            "[:up:]": Keys.ARROW_UP,
            "[:enter:]": Keys.ENTER,
        }
        split_replacements = re.split(
            "(\[:left:\]|\[:right:\]|\[:down:\]|\[:up:\]|\[:enter:\])",
            input
        )
        logger.debug(" - Split replacements input string: %s" % split_replacements)
        for chunk in split_replacements:
            if not chunk:
                continue
            if chunk in replacements:
                yield replacements[chunk]
                continue
            yield chunk

    def input(self, tag, input):
        """
        Enter some input into an element by a given tag.
        """
        logger.info(" - Injecting text \"%s\" to input" % (input))
        elem = self.element_by_tag(tag)
        self._driver_exec(self.scrolltoview, elem)
        try:
            self._driver_exec(elem.clear)
        except ElementNotInteractableException as e:
            pass
        # this handles date fields which aren't clearable
        except InvalidElementStateException as e:
            pass
        for inp in self.expand_key_substitutions(input):
            elem.send_keys(inp)
        self.path.append(("input", (tag, input,), {}))
        action = {
            "action": "input",
            "text": input,
            "tag": tag,
        }
        self.graph.add_action_to_current(action)

    def input_select_option(self, tag, option_str):
        """
        Select an input select option based on its visible string value.
        """
        logger.info(" - Selecting option %s" % (option_str))
        elem = self.element_by_tag(tag)
        self._driver_exec(self.scrolltoview, elem)
        # TODO: wrap in _driver_exec after testing
        select = Select(elem)
        # select by visible text
        select.select_by_visible_text(option_str)
        self.path.append(("input_select_option", (tag, option_str,), {}))
        action = {
            "action": "input_select_option",
            "option": option_str or "[none]",
            "tag": tag,
        }
        self.graph.add_action_to_current(action)

    def input_checkbox(self, tag, to_check, radio=False):
        """
        Check, uncheck, or don't touch an input checkbox, based
        on its current checked value.
        """
        logger.info(" - Checking to checkbox selected=%s" % (to_check))
        elem = self.element_by_tag(tag)
        self._driver_exec(self.scrolltoview, elem)
        if elem.is_selected() != to_check:
            elem.click()
        self.path.append(("input_checkbox", (tag, to_check,), {}))
        action = {
            "action": "input_checkbox",
            "checked?": to_check,
            "tag": tag,
        }
        self.graph.add_action_to_current(action)

    def click_at_position_over_element(self, elem):
        """
        Perform a "natural" click of an element, so that if there are
        any parent listeners, they will also be triggered.
        """
        logger.info(" - Performing a 'natural' click")
        from selenium.webdriver.common.action_chains import ActionChains
        ac = ActionChains(self.driver)
        x_off = 0
        y_off = 0
        ac.move_to_element(elem).move_by_offset(x_off, y_off).click().perform()

    def _text_matches_search_button(self, text):
        for test_text in ["submit", "search"]:
            if test_text == text.lower():
                return True
        return False

    def _find_submit_button(self, form):
        """
        Find a submit button by a variety of strategies:
            1. use self.form_submit_button_selector
            2. find a input with type 'submit'
            3. find a link with "search" in text
        """
        sub = None
        if self.form_submit_button_selector:
            logger.debug(" - Using submit button selector: %s" % (
                self.form_submit_button_selector
            ))
            sub = self._driver_exec(
                form.find_element_by_xpath,
                self.form_submit_button_selector
            )

        # try to find a input with type 'submit'
        if not sub:
            submit_btns = self._driver_exec(
                form.find_elements_by_xpath,
                ".//input[@type='submit']|.//button[@type='submit']"
            )
            for el in submit_btns:
                if not self.element_displayed(el):
                    continue
                txt = self.element_text(el)
                cls = el.get_attribute("class").lower()
                if self._text_matches_search_button(txt):
                    sub = el
                    logger.debug(" - Form submit input button: %s" % (txt))
                    break
                if self._text_matches_search_button(cls):
                    sub = el
                    logger.debug(" - Form submit input button (class): %s" % (cls))
                    break

        # try to find a Submit link
        # NOTE: instead of using xpath we're going to do a manual search.
        # this is due to the lack of a regex xpath matcher in some browsers
        # The below should be the functional equivaliant of something like:
        #   xpath = //a[matches(., 'submit')
        if not sub:
            try:
                possible_subs = self._driver_exec(
                    form.find_elements_by_xpath, ".//a"
                )

                for el in possible_subs:
                    if not self.element_displayed(el):
                        continue
                    el_text = self.element_text(el)
                    if not self._text_matches_search_button(el_text):
                        continue
                    sub = el
                    logger.debug(" - Form submit link: %s" % (sub))
                    break
            except NoSuchElementException as e:
                pass

        return sub

    def submit(self, tag):
        """
        Submit a form from a given tag. Assumes all inputs are filled.
        """
        logger.info("[.] Submitting form.")
        logger.debug(" - form tag: %s" % tag)
        form = self.element_by_tag(tag)
        self._driver_exec(self.scrolltoview, form)

        # try to find a Submit input button
        sub = self._find_submit_button(form)

        # attempt to click on the button, if this fails, we will
        # try to perform a form submit
        sub_failure = False
        if sub:
            try:
                if self.form_submit_natural_click:
                    self.click_at_position_over_element(sub)
                else:
                    self._loadwait(sub.click)
            except Exception as e:
                sub_failure = True
                logger.warn("Failure to click on submit button/link: %s" % e)

        # otherwise, try to submit the form itself
        if not sub or sub_failure:
            logger.debug("Using form.submit selenium shim")
            self._loadwait(form.submit, check_alerts=True)

        depth = self._get_history_depth()
        window = self._get_open_windows()[-1]
        # assuming we can't pop a new window here, for now
        self.history_stack.append([window, depth])

        self.path.append(("submit", (tag,), {}))
        node = "Submit\n tag: %s" % (tag)
        node_meta = {
            "submit": tag,
            "submit_natural_click": self.form_submit_natural_click,
        }
        self.graph.add_node(node, **node_meta)

        # TODO: better way to wait for this, post-alert clicked
        if self.form_submit_wait:
            logger.debug(
                " - Forcing post-submit wait period of %ss" %
                self.form_submit_wait
            )
            time.sleep(self.form_submit_wait)

        # move to the node if we're successful (got here, so we assume)
        self.graph.move_to_node(node)

    @property
    def page_html(self):
        """
        Get the current DOM HTML of the page.
        """
        return self._driver_exec(self.driver.page_source)

    @property
    def page_url(self):
        return self._driver_exec(self.driver.current_url)

    @property
    def infinite_loop_detected(self):
        """
        Store the recent page hashes and check to see if
        the most recent pages have all been identical. This
        determines an infinite loop. Note that this is only
        intended to be used inside a results pages loop.
        """
        infinite_loop_threshold = 5

        html_b = self.page_html
        if type(html_b) != bytes:
            html_b = self.page_html.encode("utf-8")

        hash = hashlib.md5(html_b).hexdigest()
        logger.debug(" - Page hash: %s" % (hash))

        # don't even look for infinite loop until we have the
        # required number of hashes to search
        if len(self.page_hashes) < infinite_loop_threshold:
            self.page_hashes.append(hash)
            return False

        self.page_hashes = self.page_hashes[1:] + [hash]
        recent = self.page_hashes[:infinite_loop_threshold]
        identical = recent.count(hash)
        logger.debug(" - %s pages of the past %s have been identical" % (
            identical, infinite_loop_threshold
        ))
        if identical == infinite_loop_threshold:
            return True

        return False

    def get_clickable(self, type=None):
        """
        Get tags, by type (optional), for the currently loaded page.
        """
        current_url = self._driver_exec(self.page_url)
        tagger = Tagger(
            driver=self.driver, current_url=current_url,
            leave_host=self.leave_host,
        )
        return tagger.get_clickable()

    def get_forms(self):
        current_url = self._driver_exec(self.page_url)
        tagger = Tagger(
            driver=self.driver, current_url=current_url,
            leave_host=self.leave_host,
        )
        forms_dict = tagger.get_forms()
        return forms_dict

    def get_buttons(self):
        current_url = self._driver_exec(self.page_url)
        tagger = Tagger(
            driver=self.driver, current_url=current_url,
            leave_host=self.leave_host,
        )
        return tagger.get_buttons()

    def get_screenshot(self):
        """
        Return a PNG screenshot or None if failure.
        """
        try:
            return self.driver.get_screenshot_as_png()
        # handle strange exceptions thrown from inside selenium injected
        # JS which appears to be used when taking screenshots
        except WebDriverException as e:
            logger.error("[!] Error getting screenshot: %s" % (e))
            return None



# -*- coding: UTF-8 -*-
import logging

try:
    import networkx as nx
except ModuleNotFoundError:
    pass


logger = logging.getLogger('AUTOSCRAPE')


class Graph(object):
    def __init__(self):
        try:
            self.graph = nx.DiGraph()
        except NameError:
            logger.debug(
                "NetworkX not installed. Not building crawl graph."
                " (Hint: pip install autoscrape[graph])"
            )
            self.graph = None
        # store scrape graph
        # current node (css tag)
        self.current = None

    def add_root_node(self, node, **kwargs):
        """
        Add an isolated node, make it the current node. This
        is for creating new subgraphs inside our graph. Does
        not add any edges.
        """
        if self.graph is None:
            return
        self.graph.add_node(node, **kwargs)
        self.current = node

    def add_node(self, node, **kwargs):
        """
        Add a single node to the tree, with edges connecting
        to the current node.
        """
        if self.graph is None:
            return
        self.graph.add_node(node, **kwargs)
        self.graph.add_edge(self.current, node)

    def add_nodes(self, nodes):
        """
        Add a list of nodes to the current node in the graph.
        This handles adding the nodes and the edges.
        """
        if self.graph is None:
            return
        for node, meta in nodes:
            self.add_node(node, **meta)

    def add_meta_to_current(self, **meta):
        if self.graph is None:
            return
        self.graph.nodes[self.current].update(**meta)

    def add_action_to_current(self, action):
        if self.graph is None:
            return
        current_meta = self.graph.nodes[self.current]
        current_actions = current_meta.get("actions", [])
        current_actions.append(action)
        current_meta["actions"] = current_actions
        nx.set_node_attributes(self.graph, current_meta, name=self.current)

    def move_to_node(self, node):
        if self.graph is None:
            return
        self.current = node

    def move_to_parent(self):
        if self.graph is None:
            return
        try:
            preds = self.graph.predecessors(self.current)
            parent = self.graph.predecessors(self.current).__next__()
        except StopIteration:
            return
        self.move_to_node(parent)

    def save_graph(self, output_path):
        if self.graph is None:
            return
        nx.write_gpickle(self.graph, output_path)



# -*- coding: UTF-8 -*-
import base64
import json
import logging
import os
import re
from urllib import parse

import requests


logger = logging.getLogger('AUTOSCRAPE')


def get_filename_from_url(url):
    """
    Take a fully-qualified URL and turn it into a filename. For
    example, turn a url like this:

        https://www.cia.gov/library/readingroom/docs/%5B15423241%5D.pdf

    Using the parsed URL:

        ParseResult(scheme='https', netloc='www.cia.gov',
            path='/library/readingroom/docs/%5B15423241%5D.pdf

    Returing this representation (a string):

        _library_readingroom_docs_%5B15423241%5D.pdf

    NOTE: If no extension is found on the page, .html is appended.
    """
    parsed = parse.urlparse(url)
    host = parsed.netloc
    # split filename/path and extension
    file_parts = os.path.splitext(parsed.path)
    # file_part = file_parts[0].replace("/", "__")
    extension = file_parts[1] or ".html"
    # filename = "%s_%s" % (host, file_part)
    filename = url.replace("/", "_").replace(":", "_")
    if parsed.query:
        query_part = "_".join(parsed.query.split("&"))
        filename = "%s__%s" % (filename, query_part)
    return "%s%s" % (filename, extension)


def get_extension_from_url(url):
    # try and extract the extension from the URL
    path = parse.urlparse(url).path
    ext = os.path.splitext(path)[1]
    ext = ext if ext else "html"
    if ext[0] == ".":
        ext = ext[1:]
    return ext


def write_file(filepath, data, fileclass=None, writetype="w", output=None,
               url=None):
    """
    Write out a scraped data file to disk or a remote callback,
    specified in output parameter.
    """
    logger.debug("[.] Writing file: %s to: %s" % (filepath, output))
    if not output:
        return

    # Rest API callback mode
    if re.match("^https?://", output):
        # (b64encode) bytes -> (decode) str
        if type(data) == bytes:
            encoded = base64.b64encode(data).decode()
        else:
            encoded = base64.b64encode(bytes(data, "utf-8")).decode()
        payload = {
            "name": filepath,
            "data": encoded,
            "fileclass": fileclass,
            "url": url,
        }
        if fileclass:
            payload["fileclass"] = fileclass
            post_data = json.dumps(payload).encode("utf-8")
            headers = {
                "content-type": "application/json"
            }
            r = requests.post(
                output, data=post_data, headers=headers
            )
            r.status_code

    # filesystem mode
    else:
        dirpath = os.path.dirname(filepath)
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        with open(filepath, writetype) as f:
            f.write(data)
